# Sulk - main

#*****************************************************************************
#
#                                   Sulk
#
#       Copyright (c) 2002 Toby Woodwark <zirtix@softhome.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2, or 
# at your option any later version, as published by the Free Software 
# Foundation (see COPYING).
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# FSF website: http://www.fsf.org
#
#******************************************************************************
#
# 'Space Hulk' is a trademark of Games Workshop Limited.  For further 
# copyright, trademark and authorship information see README.
#
#******************************************************************************


"""An implementation of the board game 'Space Hulk', using Pygame.

Sulk aims to be as complete as possible a simulation of the first edition game.
It has the marine clock, CPs in the stealer turn, and all the other classic
features.  To understand some parts of this code, you have to know the rules of
Space Hulk.  Sorry!
"""

# See HACKING for info on code conventions, TODO and so on


##############################################################################
####### Import / set up

## Standard modules + pygame
from __future__ import generators #req Python 2.2, (so we can use bools too)
import pygame, random, gc
from pygame.locals import *

## Sulk modules
# Non-variable ints, Rects, fonts, colors; plus config options 
from constants import *
# For loading images, etc
import util
# Crude global sound control
import sound
# Graphical objects
import sprites
from sprites import Graphic
# Miscellaneous classes, sprite groups
import objects
# Mission modules from MISSIONS_DIRECTORY
import missions
# Game sequence and UI objects
import players


##############################################################################
####### Sulkgame object

class SulkGame:
    """The main class.
    
    Instantiated each time there is a new game.
    
    Attributes (some may also be set to None):
    (instance)
    logic--
	players     -- tuple of two player modules.  First elt is always
                       the local player module, players.local
	cplayer	    -- current player instance
	mission	    -- module
	paused	    -- bool
	_pause_to_clear -- bool -- if true we need to clear the next 'elapsed'
				    count at unpause 
	_done	    -- int -- if true, we exit
	phase	    -- Phase instance
	_blip_list  -- list of ints -- representing a deck of blips
    graphics--
	board	    -- objects.board.Board instance -- game board
	donebutton  -- sprites.DoneButton instance  (FIXME?)
	cpbutton    -- sprites.CPButton instance    (FIXME?)
	cpclock	    -- sprites.CPClock instance	    (FIXME?)
	littlemapbox-- sprites.LittleMapBox instance
    SpriteGroups--
	_last_actor -- GroupSingle -- the last piece to do something (TODO -- to Phase?)
        _oldpiece   -- Sprite or Null (CRUFT)
	stealers
	marines
	yada yada...
	"""

    def __init__(self):

	# Get a layer array (render groups)
	self.layer_array = objects.rendering.LayerArray(
        sprites.Clock,
        sprites.SelectDec, 
        sprites.Marker,
        sprites.MapFade,
        sprites.SectionLine, 
        sprites.HiLite, 
        sprites.SquareComment, 
        sprites.FacingDec,
        sprites.Button, 
        sprites.TurnTimer, 
        sprites.GlobalButton,
        sprites.MiniSprite, 
        sprites.LittleMapBox,
        sprites.Feature, 
        sprites.Flames, 
        sprites.Flash,
        sprites.C_A_T,
        sprites.EntryTriangle, 
        sprites.MiniSquare,
        sprites.ScrollMessage, 
        sprites.ToolTip,
        sprites.NewDialog, 
        sprites.Square,
        sprites.Piece,
        )

	# Nonrendering sprite groups and their extra attributes
        self.stealers	    = objects.teams.StealerTeam(self)
	self.marines	    = objects.teams.MarineTeam(self)
	self.neutrals	    = objects.teams.NeutralTeam(self)
        self.marines.enemy_team	    = self.stealers
        self.stealers.enemy_team    = self.marines

	# Containers - TODO -- get rid of some
    	sprites.Feature.all_features	= \
	self.features			= pygame.sprite.Group()
        self.pieces = pygame.sprite.Group()
	sprites.Square.all_squares	= \
	self.squares			= pygame.sprite.Group()
	sprites.LimboSquare.all_limbosquares = \
	self.limbosquares		= pygame.sprite.Group()
	sprites.EntryTriangle.all	= \
	self.entrytriangles		= pygame.sprite.Group()
	sprites.ExitArrow.all		= \
	self.exitarrows			= pygame.sprite.Group()
	sprites.Button.all_buttons	= \
	self.buttons			= pygame.sprite.Group()
	sprites.DisplayDie.most_recent	= None # reset
        sprites.NewDialog.all_dialogs   = \
	self.dialogs = sprites.NewDialog.render_container #hmm
	sprites.Terminator_Marine.team		= self.marines
	sprites.Genestealer.team		= self.stealers
	sprites.C_A_T.team		= self.neutrals

	# Set attributes
	self.tooltip_wait = TOOLTIP_TIME
	self.tooltip_kill_wait = TOOLTIP_KILL_TIME
	self.hoversprite = None
	self.tooltip = None
	self.fpsclock = None
	self.paused = False
	self._pause_to_clear = False
	self._done = False
	self._last_actor = pygame.sprite.GroupSingle()
        self.mission = None
        
	# Set phase up -- get self.marines.player, self.mission etc
        self.teams = {MARINETEAM:self.marines, STEALERTEAM:self.stealers} #hmm
	players.local.phases.Choose_Players(self) #new Phase object (leads on to mission choice)
        
	while not self.mission: 
	    #N.B. calling _game_loop() early could cause problems with InputHandler
	    if self._done: return
	    self._game_loop()
            
        print >> DEBUG_STREAM,\
        "SulkGame --> init, marine and stealer players:", self.marines.player, self.stealers.player #DEBUG
        print >> DEBUG_STREAM, "SulkGame --> init, self.mission:", self.mission #DEBUG

	# Add an intro message and stuff
	#sprites.MapFade(1) #this is lame - TODO replace with blinds or some other anim
	sprites.ScrollMessage("Hello, alpha testers!")
	
	# Buttons -- TODO remove these references, make them attach to phase (?)
	self.globalbuttons = \
	self.donebutton, self.cpbutton, \
	self.pausebutton, self.secretsbutton = \
	sprites.DoneButton(self), sprites.CPButton(self), \
	sprites.PauseButton(self), sprites.SecretsButton(self)
	
	# Set up blips
	self._blip_list = []
	for key in self.mission.BLIP_SET.keys():
	    for x in range(self.mission.BLIP_SET[key]):
		self._blip_list.append(key)
		
	# Make board, squares, yada yada
	self.board = objects.board.Board(self)

	# Set up nav box
	self.littlemapbox = sprites.LittleMapBox(self.board)
	
	# Add Pregame
        self.mission.init(self)
	players.local.phases.Pregame_Phase(self)


    def get_blipval(self):
	"""get_blipval() -> int
	get a random blip value"""
	
	# RULES 
        # Should we attempt to emulate the blip stack here (biasing the 
	# probabilities and allowing 'card-counting')?
	
	return random.choice(self._blip_list)
    
    
    def get_ambush_counter_val(self):
        """get_ambush_counter_val(self) -> int
	get 1 (for a stealer) or 0 (false alarm)"""
        
        #1/3 chance (RULES?)
        return random.choice((0, 0, 1))

    
    def set_actor(self, piece=None):
	"""set_actor(piece) -> None
	set the active piece
	
	Set or unset the piece that acted last."""
	
	if piece and piece.alive():
	    piece.add(self._last_actor)
	else:
	    self._last_actor.empty()


    def get_actor(self):
	"""get_actor() -> Piece instance / None
	get the active piece
	
	Returns the piece that last completed an action."""
	
	if self._last_actor:
	    return self._last_actor.sprite #N.B. GroupSingle internals - bad!  -TW
    
 
    def run_game(self):
	"""run_game() -> None
	keep looping
	
	Loop until the game's over, or we want to play a different mish"""
	
	while not (another_game or self._done):
	    self._game_loop()


    def kill_tooltip(self):
	"""kill_tooltip() -> None
	remove any tooltip there is"""
	
	if self.tooltip: 
	    self.tooltip.kill()
	    self.tooltip = None
	    self.tooltip_kill_wait = TOOLTIP_KILL_TIME
	    self.tooltip_wait = TOOLTIP_TIME

    
    def toggle_fpsclock(self):
	"""toggle_fpsclock() -> None
	toggle FPS display"""
	
	if self.fpsclock: 
	    self.fpsclock.kill()
	    self.fpsclock = None
	else:
	    self.fpsclock = sprites.FPSClock()
	    
	
    def pause(self):
	"""pause() -> bool
	pause the game

	Returns true if game is already paused or becomes paused.
	Returns false if it makes no sense to be paused.
	
	Pause stops animated things from animating, and stops the clock."""
	
	if self.paused: 
	    print >> DEBUG_STREAM, "SulkGame --> pause, game already paused!" #DEBUG
	    return True
	elif not self.phase.pausable:
	    print >> DEBUG_STREAM, "SulkGame --> pause, phase not pausable!" #DEBUG
	    return False
	
	self.paused = True
	self._pause_to_clear = True

	# Kill tooltip, if any
	self.kill_tooltip()

	# Stop player using piece
	self._oldpiece = getattr(self.phase, 'cpiece', util.Null)() #piece or Null
	self.phase.unselect()
	
	return True


    def unpause(self):
	"""unpause() -> None
	unpause the game"""
	
	if not self.paused: 
	    print >> DEBUG_STREAM, "SulkGame --> unpause, game is not paused!" #DEBUG
	    return
	
	self.paused = False

	# Allow play
	if self._oldpiece.alive(): 
	    self.phase.select(self._oldpiece)
	del self._oldpiece


    def _game_loop(self):
	"""_game_loop() -> None
	render and handle input once

	Do one loop of rendering and event queue handling for the game.
	Use dirty rect animation to draw everything onto screen."""
	
	# Clear to background
	self.layer_array.clear_all(screen, background)
	
	# Call game logic / handlers
	self.phase.loop()
	
        if not self.paused:
	    # Tick
	    elapsed = clock.tick(MAX_TICK_RATE)
	    
	    if self._pause_to_clear:
		self._pause_to_clear = False
		Graphic.elapsed = 0
		Graphic.speed_ratio = 0.0
	    else:
		# Send elapsed time & fps to Graphic class
		Graphic.elapsed = elapsed
		Graphic.speed_ratio = float(elapsed) / MIN_TICK_LENGTH #1.0 to inf
		Graphic.fps = clock.get_fps()

		# Do tooltip stuff
		if 1 in pygame.mouse.get_pressed(): #hmm, kludgey
		    # Someone's clicking the mouse
		    self.kill_tooltip()
		else:
		    # Seek for tooltip
		    s = self.layer_array.get_TT_sprite_at(pygame.mouse.get_pos())
		    # The sprite must exist, and return a nontrivial tooltip string
		    if s and s.tt_text():
			if s is self.hoversprite:
			    # We have hovered here before
			    if not self.tooltip:
				# Timer runs down
				self.tooltip_wait -= elapsed
				if self.tooltip_wait <= 0:
				    # Make a tooltip
				    self.tooltip_wait = TOOLTIP_TIME
				    self.tooltip = sprites.ToolTip(s)
			elif self.tooltip:
			    # 'Switch tooltips'
			    self.kill_tooltip()
			    self.tooltip = sprites.ToolTip(s)
		    # We couldn't find a tooltip-able sprite
		    elif self.tooltip:
			# Timer runs down
			self.tooltip_kill_wait -= elapsed
			if self.tooltip_kill_wait <= 0:
			    self.kill_tooltip()
		    self.hoversprite = s
		    
	    # Update some sprites
	    self.layer_array.update_autos()
	    
	# Update display object and draw    
	pygame.display.update(self.layer_array.draw_all(screen))


    def game_over_check(self):
	"""game_over_check() -> bool
	see if the game is over
	
	If the game is indeed over, set up some gfx, change self.phase, and
	return True."""
	
	victoriousteam = self.mission.victory_check(self)
	if victoriousteam is not None:
	    for team in (self.stealers, self.marines):
		if team.casualties == 1:
		    sprites.ScrollMessage("The %s took one casualty."%\
		    team.type)
		elif team.casualties == 0:
		    sprites.ScrollMessage("The %s took no casualties."%\
		    team.type)
		else:
		    sprites.ScrollMessage("The %s took %d casualties."%\
		    (team.type, team.casualties))
	    if victoriousteam == NEITHER: 
		text = "The game is a draw."
	    else: 
		text = "The %s win!" % victoriousteam.type
	    players.local.phases.Postgame_Phase(self, text)
	    return True
	
	return False


    def clear_screen(self):
	"""clear_screen() -> None
	prep and blit background
	
	Must be called before SulkGame.__init__() draws to the background."""
	#FIXME this does not need to be a method of SulkGame
	
	background.fill(BG_COLOR)
	for r in (MAP_RECT, SCROLLMESSAGE_RECT, MINIMAP_RECT):
	    background.fill(MAP_BG_COLOR, r.inflate(2,2))
	for r in (MAP_RECT, SCROLLMESSAGE_RECT, MINIMAP_RECT, PIECEBUTTONS_RECT, TURNTIMER_RECT):
	    pygame.draw.rect(background, RECT_COLOR, r.inflate(2,2), 1)
	screen.blit(background, (0,0))


    def end(self):
	"""end() -> None
	end the current mission and exit"""
	#TODO more gc / tidying up?
	#TODO a verification dialog
	
	self._done = True


    def restart(self):
	"""end() -> None
	end the current mission
	
	Triggers another call of run_one()"""
	
	global another_game
	another_game = True
	self.end()


##############################################################################
####### Data

def load_data():
    """load_data() -> None
    
    Loads all data for the classes imported from 'sprites'.
    Displays a splash logo and IO feedback."""
    
    # There are this many things to load. (Must match no. of load_x(...) calls!) FIXME
    number_pending = 115
    
    # Put up the logo    
    splash = util.load.image('splash.png', TOPLEFT_COLORKEY) 
    splash_rect = splash.get_rect()
    splash_rect.midtop = SPLASH_MIDTOP
    screen.blit(splash, splash_rect)

    # Put up feedback msg
    mesg = COMMENT_FONT.render("version %s loading..."%VERSION_STRING, 1, GREY1)
    mesg_rect = mesg.get_rect()
    mesg_rect.top = splash_rect.bottom + PAD
    mesg_rect.centerx = splash_rect.centerx
    screen.blit(mesg, mesg_rect)

    # Prep for feedback routine
    width = int(MAP_RECT.width / number_pending)
    feedback_rect = Rect(0, mesg_rect.bottom + (2 * PAD),
			number_pending * width, 2 * PAD)
    feedback_rect.centerx = splash_rect.centerx
    pygame.draw.rect(screen, GREY1, feedback_rect, 1)
    
    # Update visual
    pygame.display.update((feedback_rect, mesg_rect, splash_rect))

    # Define feedback routine
    def fbgen():
	"""GENERATOR
	next() Blits to and updates feedback rect"""
        
	feedback_left = feedback_rect.left
	feedback_top = feedback_rect.top
	feedback = pygame.Surface((width, 2 * PAD))
	feedback.fill(GREY2)
	alphincrement = int(130/number_pending)
	i = 0
	while i < number_pending:
	    feedback.set_alpha(125 + i * alphincrement)
	    screen.blit(feedback, (feedback_left + (i * width), feedback_top))
	    i += 1
	    yield pygame.display.update(feedback_rect)

    fb = fbgen()

    def fbnext():
        try:
            fb.next()
        except StopIteration:
            pass #?

    # Shortcuts
    def load_t_img(*args):
	"load a themed image"
	fbnext()
	return util.load.theme_image(*args)
    
    def load_img(*args):
	"load an image"
	fbnext()
	return util.load.image(*args)
    
    def load_snd(*args):
	"load a sound"
	fbnext()
	s = util.load.sound(*args)
        sound.all_sounds.append(s)
        return s

    #####Pieces#####
    sprites.C_A_T.image	=	    load_t_img('cat.png',    PERPIXEL_ALPHA)
    sprites.Genestealer.image =	    load_t_img('stealer.png',PERPIXEL_ALPHA)
    sprites.Genestealer.mini_image  =   load_t_img('mini_stealer.png')
    sprites.Blip.image =	    load_t_img('blip.png',   PERPIXEL_ALPHA)
    sprites.Blip.revealedimg = {1:  load_t_img('blip_1.png', PERPIXEL_ALPHA), 
				2:  load_t_img('blip_2.png', PERPIXEL_ALPHA), 
				3:  load_t_img('blip_3.png', PERPIXEL_ALPHA)}
    sprites.Ambush_Counter.image =  load_t_img('ambush_counter.png',   PERPIXEL_ALPHA)
    sprites.Blip.mini_image  =	    load_t_img('mini_blip.png')
    sprites.Terminator_Marine.image =	                load_t_img('terminator_storm_bolter.png', PERPIXEL_ALPHA)
    sprites.Terminator_Marine.mini_image =	        load_t_img('mini_terminator.png')
    sprites.Terminator_Marine_Sergeant.image =	        load_t_img('terminator_sergeant.png',  PERPIXEL_ALPHA)
    sprites.Terminator_Marine_Sergeant_with_Power_Sword.image =	        load_t_img('terminator_sergeant_sword.png',  PERPIXEL_ALPHA)    
    sprites.Terminator_Marine_with_Heavy_Flamer.image =	load_t_img('terminator_heavy_flamer.png', PERPIXEL_ALPHA)
    sprites.Terminator_Marine_with_Lightning_Claws.image =	    load_t_img('terminator_lightning_claws.png', PERPIXEL_ALPHA)
    sprites.Terminator_Marine_with_Assault_Cannon.image =	    load_t_img('terminator_assault_cannon.png', PERPIXEL_ALPHA)
    sprites.Terminator_Marine_with_Thunder_Hammer.image =   load_t_img('terminator_thunder_hammer.png', PERPIXEL_ALPHA)
    sprites.Terminator_Marine_with_Chain_Fist.image =   load_t_img('terminator_chain_fist.png', PERPIXEL_ALPHA)
    sprites.Terminator_Marine_Captain.image =           load_t_img('terminator_captain.png', PERPIXEL_ALPHA)
    sprites.Terminator_Marine_Librarian.image =         load_t_img('terminator_librarian.png',PERPIXEL_ALPHA)
    
    #####Misc stuff#####
    sprites.Square.corridor_image = load_t_img('square_corridor.png')
    sprites.Square.room_image =	    load_t_img('square_room.png')
    sprites.LimboSquare.image =	    util.zerosurface.convert() # zero sized non-image
    sprites.EntryTriangle.image =   load_t_img('entry.png',	PERPIXEL_ALPHA)
    sprites.ExitArrow.image =	    load_t_img('exit.png',	PERPIXEL_ALPHA)
    sprites.SectionLine.image =	    load_t_img('sectionline.png',TOPLEFT_COLORKEY)
    sprites.Ducting.image =	    load_t_img('ducting.png',	PERPIXEL_ALPHA)
    sprites.Ducting.destroyed_image =load_t_img('ducting_destroyed.png',PERPIXEL_ALPHA)
    sprites.Door.image =	    load_t_img('door_closed.png',PERPIXEL_ALPHA)
    sprites.Door.openimage =	    load_t_img('door_open.png', PERPIXEL_ALPHA)
    sprites.Bulkhead.image =        load_t_img('bulkhead.png', PERPIXEL_ALPHA)
    sprites.Bulkhead.destroyed_image = load_t_img('bulkhead_destroyed.png', PERPIXEL_ALPHA)
    sprites.Cryogenic_Tank.image =  load_t_img('cryogenic_tank.png', PERPIXEL_ALPHA)
    sprites.Cryogenic_Tank.destroyed_image =   load_t_img('cryogenic_tank_destroyed.png', PERPIXEL_ALPHA)
    sprites.Crate.image =           load_t_img('crate.png',PERPIXEL_ALPHA)
    sprites.Rubble.image =          load_t_img('rubble.png',PERPIXEL_ALPHA)
    sprites.Flames.image =	    load_t_img('flames.png',	PERPIXEL_ALPHA)
    sprites.Flash.images_map = {
    GUN:	[                   load_t_img('flash_storm_bolter.png',PERPIXEL_ALPHA), 
		                    util.zerosurface.convert()
                ],
    FLAMEGUN:	[                   load_t_img('flash_heavy_flamer.png',PERPIXEL_ALPHA)
                ],
    ASSAULT_CANNON: [               util.zerosurface.convert(),
                                    load_t_img('flash_heavy_flamer.png',PERPIXEL_ALPHA) #FIXME
                ]
                                }
    
    #####Markers#####
    sprites.OWMarker.image =	    load_t_img   ('marker_overwatch.png',   PERPIXEL_ALPHA)
    sprites.JamMarker.image =	    load_t_img   ('marker_jam.png',	    PERPIXEL_ALPHA)
    sprites.ConvertMarker.img = {   1:load_t_img ('marker_convert_1.png',   PERPIXEL_ALPHA),
				    2:load_t_img ('marker_convert_2.png',   PERPIXEL_ALPHA),
				    3:load_t_img ('marker_convert_3.png',   PERPIXEL_ALPHA)
                                }
    sprites.DamageMarker.image =    load_t_img   ('marker_damage.png',	    PERPIXEL_ALPHA)

    #####UI elements#####
    sprites.SelectDec.image =	    load_t_img('select.png',	    TOPLEFT_COLORKEY)
    sprites.SelectDec.cat_image =   load_t_img('select_cat.png',    TOPLEFT_COLORKEY)
    sprites.FacingDec.image =	    load_t_img('facing_arrow.png',  TOPLEFT_COLORKEY)
    # Hilite images are constructed at load time:
    redimg = pygame.Surface((SQ_LENGTH, SQ_LENGTH)).convert()
    redimg.fill(HILITE_RED_COLOR)
    pygame.draw.rect(redimg, HILITE_RED_BORDER_COLOR, redimg.get_rect(), 3) #hmm
    redimg.set_alpha(HILITE_ALPHA)
    greenimg = pygame.Surface((SQ_LENGTH, SQ_LENGTH)).convert()
    greenimg.fill(HILITE_GREEN_COLOR)
    greenimg.set_alpha(HILITE_ALPHA)
    # End of image making
    sprites.HiLite.img = {	    REDHILITE:redimg,
				    GREENHILITE:greenimg}
    sprites.MiniSquare.image  =	    load_t_img('mini_square.png')
    sprites.MiniLimboSquare.image = util.zerosurface.convert()
    sprites.MiniEntryExit.image = load_t_img('mini_entryexit.png')
    
    #####Buttons#####
    sprites.CheckboxWidget.boximg = {
    PRESSED :load_img('checkbox_checked.png',	PERPIXEL_ALPHA), #DISABLED?
    ENABLED :load_img('checkbox_empty.png',	PERPIXEL_ALPHA)}
    sprites.RadioButtonWidget.boximg = {
    PRESSED :load_img('radiobox_checked.png',	PERPIXEL_ALPHA),
    ENABLED :load_img('radiobox_empty.png',	PERPIXEL_ALPHA)}
    sprites.DoneButton.img = {	
    DISABLED:load_img('done_disabled.png',	PERPIXEL_ALPHA),
    PRESSED :load_img('done_pressed.png',	PERPIXEL_ALPHA),
    ENABLED :load_img('done_enabled.png',	PERPIXEL_ALPHA)}
    sprites.SecretsButton.img ={
    DISABLED:load_img('secrets_disabled.png',	PERPIXEL_ALPHA),
    PRESSED :load_img('secrets_pressed.png',	PERPIXEL_ALPHA),
    ENABLED :load_img('secrets_enabled.png',	PERPIXEL_ALPHA)}
    sprites.PauseButton.img = {
    DISABLED:load_img('pause_disabled.png',	PERPIXEL_ALPHA),
    PRESSED :load_img('pause_pressed.png',	PERPIXEL_ALPHA),
    ENABLED :load_img('pause_enabled.png',	PERPIXEL_ALPHA)}
    sprites.CPButton.img = {
    DISABLED:load_img('cp_disabled.png',	PERPIXEL_ALPHA),
    PRESSED :load_img('cp_pressed.png',		PERPIXEL_ALPHA),
    ENABLED :load_img('cp_enabled.png',		PERPIXEL_ALPHA)}
    sprites.AttackButton.img = {
    ENABLED :load_img('attack_enabled.png',	PERPIXEL_ALPHA),
    DISABLED:load_img('attack_disabled.png',	PERPIXEL_ALPHA)}
    sprites.AutofireButton.img = {
    ENABLED :load_img('attack_enabled.png',	PERPIXEL_ALPHA),
    DISABLED:load_img('attack_disabled.png',	PERPIXEL_ALPHA)}
    sprites.CutDoorButton.img = {
    ENABLED :load_img('attack_enabled.png',	PERPIXEL_ALPHA),
    DISABLED:load_img('attack_disabled.png',	PERPIXEL_ALPHA),
    PRESSED:load_img('usedoor_pressed.png',     PERPIXEL_ALPHA),
    }
    sprites.UnjamButton.img = {
    ENABLED :load_img('unjam_enabled.png',	PERPIXEL_ALPHA),
    DISABLED:load_img('unjam_disabled.png',	PERPIXEL_ALPHA)}
    sprites.ShootButton.img = {
    ENABLED :load_img('shoot_enabled.png',	PERPIXEL_ALPHA),
    PRESSED :load_img('shoot_pressed.png',	PERPIXEL_ALPHA),
    DISABLED:load_img('shoot_disabled.png',	PERPIXEL_ALPHA)}
    sprites.FireGrenadeButton.img = {
    ENABLED :load_img('shoot_enabled.png',	PERPIXEL_ALPHA),
    PRESSED :load_img('shoot_pressed.png',	PERPIXEL_ALPHA),
    DISABLED:load_img('shoot_disabled.png',	PERPIXEL_ALPHA)}
    sprites.SelfdestructButton.img = {	
    ENABLED :load_img('selfdestruct_enabled.png',PERPIXEL_ALPHA),
    PRESSED :load_img('selfdestruct_pressed.png',PERPIXEL_ALPHA),
    DISABLED:load_img('selfdestruct_disabled.png',PERPIXEL_ALPHA)}
    sprites.OverwatchButton.img = {	
    ENABLED :load_img('overwatch_enabled.png',  PERPIXEL_ALPHA),
    PRESSED :load_img('overwatch_pressed.png',  PERPIXEL_ALPHA),
    DISABLED:load_img('overwatch_disabled.png', PERPIXEL_ALPHA)}
    sprites.MovenshootButton.img = {
    DISABLED:load_img('movenshoot_disabled.png',PERPIXEL_ALPHA),
    ENABLED :load_img('movenshoot_enabled.png', PERPIXEL_ALPHA),
    PRESSED :load_img('movenshoot_pressed.png', PERPIXEL_ALPHA)}
    sprites.UsedoorButton.img = {
    ENABLED :load_img('usedoor_enabled.png',    PERPIXEL_ALPHA),
    PRESSED :load_img('usedoor_pressed.png',    PERPIXEL_ALPHA),
    DISABLED:load_img('usedoor_disabled.png',   PERPIXEL_ALPHA)}
    sprites.ConvertButton.img = {
    DISABLED:load_img('convert_disabled.png',   PERPIXEL_ALPHA),
    ENABLED :load_img('convert_enabled.png',    PERPIXEL_ALPHA)}
    sprites.CATButton.img = {
    DISABLED:load_img('cat_disabled.png',	PERPIXEL_ALPHA),
    ENABLED :load_img('cat_enabled.png',	PERPIXEL_ALPHA),
    PRESSED :load_img('cat_pressed.png',	PERPIXEL_ALPHA)}
    sprites.ReloadButton.img = {
    DISABLED:load_img('cat_disabled.png',	PERPIXEL_ALPHA),
    ENABLED :load_img('cat_enabled.png',	PERPIXEL_ALPHA),
    PRESSED :load_img('cat_pressed.png',	PERPIXEL_ALPHA)}

    #####Sounds#####
    #TODO -- Sounds for more events
    #	    + start of game
    #	    + timer start and timeout
    #	    + marines interrupt
    #	    + CAT, esp for damage() and kill()
    #	    + unjamming
    #	    + blip conversion
    sprites.Flash.sound_map = {
    GUN:                            load_snd('marine_shoot_bolter.wav'), #OK (TW)
    FLAMEGUN:                       load_snd('marine_shoot_flamer.wav'), #OK (TW)
    ASSAULT_CANNON:                 load_snd('assault_cannon_burst.wav')} #legally dodgy (TW)
    sprites.Button.fail_sound =	    load_snd('button_fail.wav')
    sprites.Button.press_sound =    load_snd('button_press.wav')
    sprites.Button.unpress_sound =  load_snd('button_unpress.wav')
    snd_punch =			    load_snd('marine_cc.wav')
    snd_clang =			    load_snd('marine_move.wav')
    snd_scream =		    load_snd('marine_kill_skewered.wav')
    sprites.Terminator_Marine.movesound = \
    sprites.Terminator_Marine.turnsound =   snd_clang #OK, but turnsound should differ (TW)
    sprites.Terminator_Marine.diesound =    snd_scream #yes (TW)
    sprites.Terminator_Marine.ccsound =     snd_punch #maybe (TW)
    sprites.Terminator_Marine.jamsound =    load_snd('marine_jam.wav')#no (TW)
    sprites.Terminator_Marine_with_Heavy_Flamer.destructsound =  \
    sprites.Terminator_Marine_with_Thunder_Hammer.destructsound = load_snd('marine_selfdestruct_flamer.wav') #good (TW)
    sprites.Terminator_Marine_with_Chain_Fist.cc_sound =    load_snd('chain_fist.wav')
    sprites.Blip.movesound = \
    sprites.Genestealer.movesound = \
    sprites.Genestealer.turnsound =     snd_clang #no (TW)
    #sprites.Blip.convertsound =    load_snd('nofile.wav') TODO
    sprites.Blip.diesound = \
    sprites.Genestealer.diesound =      snd_scream #no (TW)
    sprites.Genestealer.ccsound =       snd_punch #no (TW)
    sprites.Door.opensound = \
    sprites.Door.closesound =       load_snd('door_open.wav') #I like this one (TW)
    sprites.Widget.click_sound =    load_snd('click.wav') #this sucks, too quiet (TW)

    # Supply sounds to the sound module
    sound.test_sound = snd_clang


##############################################################################
####### Main functions
def run_one():
    """run_one() -> None
    
    Run a game and return."""
    
    print >> DEBUG_STREAM, "	    -- Game of %s begins. --"%APP_AND_VERSION_STRING #DEBUG
    
    # We don't want to play again by default
    global another_game
    another_game = False

    # Set up a sulkgame object
    newgame = SulkGame()
        
    # Loop until new mission chosen/restart
    newgame.run_game()
    
    # Game over
    print >> DEBUG_STREAM, "	    -- Game has ended. --" #DEBUG
    

def main():  
    """main() -> None"""
    
    global screen, background, clock, another_game

    another_game = True
    
    # Initialize pygame
    pygame.init()
    random.seed()
    pygame.key.set_repeat(400, 45)
    pygame.event.set_allowed(None)
    pygame.event.set_allowed((QUIT, MOUSEBUTTONDOWN, MOUSEBUTTONUP, KEYDOWN)) #hmm

    # Prepare screen
    if options.SCREEN_FULLSCREEN: 
	screen = pygame.display.set_mode(SCREEN_SIZE, FULLSCREEN)
    else: 
	screen = pygame.display.set_mode(SCREEN_SIZE)
    
    # Print some info
    print >> DEBUG_STREAM, "Pygame version:", pygame.ver #DEBUG
    print >> DEBUG_STREAM, "display driver:", pygame.display.get_driver()#DEBUG
    print >> DEBUG_STREAM, "display info:", pygame.display.Info()#DEBUG
    
    # Icon
    pygame.display.set_icon(util.load.image('icon.png'))
    
    # Screen caption
    pygame.display.set_caption(APP_AND_VERSION_STRING)
    
    # Objects
    background = pygame.Surface(SCREEN_SIZE)
    clock = pygame.time.Clock()
    
    # Data
    load_data()
    
    # Set soundvol to options level
    sound.set_vol(sound.get_vol())
    
    # Check if sound wanted
    if not pygame.mixer.get_init():
        options.SOUND_ENABLED = False
    elif not options.SOUND_ENABLED:
	pygame.mixer.quit() #KLUDGE
	
    # Brief wait
    util.hang(0.7) #remove if too annoying

    # Run games
    while another_game:	
	# Calling run_one() >once leaks memory through the sulkgame object, hence gc
	run_one()
	print >> DEBUG_STREAM, "main, gc.collect:", gc.collect() #DEBUG
