# Sulk - objects - rendering

#*****************************************************************************
#
#                                   Sulk
#
#       Copyright (c) 2002 Toby Woodwark <zirtix@softhome.net>
#
# This program is free software.  See the file COPYING for details.
#
#*****************************************************************************

"""
Sulk graphic group and class objects.
"""

##############################################################################
####### Import

import pygame

from constants import DEBUG_STREAM


##############################################################################
####### Classes

class LayerArray:
    """Class to build, contain and organise ClipGroup instances.
    
    Generates clgroups to order based on Graphic subclass attrs.
    
    Attributes:
    (instance)
	clgroups -- a list of groups -- in rendering order
	"""

    def __init__(self, *classes):
	
	# N.B. Each class in *classes should be a subclass of sprites.Graphic
	self.clgroups = []
	self._upgroups = []
	
	for cl in classes:
	    #print >> DEBUG_STREAM, "LayerArray --> init --> cl:", cl#DEBUG
	    self._configure_class(cl)
	    
	self._sort_clgroups()

    
    def _sort_clgroups(self):
	"""-> None"""
	
	def cmpfunc(clgp1, clgp2):
	    """-> -1 or 0 or 1

	    clgp1 > clgp2 means clgp1 should be drawn *later* than clgp2.
	    We want higher layers and bigger-clipped sprites drawn later."""

	    if clgp1.layer > clgp2.layer: return 1
	    elif clgp1.layer < clgp2.layer: return -1
	    else:
		if clgp1.clip is clgp2.clip: return 0
		elif clgp2.clip is None: return -1
		elif clgp1.clip is None or clgp1.clip.contains(clgp2.clip): return 1
		else: return -1
		
	self.clgroups.sort(cmpfunc)


    def _configure_class(self, cl):
	"""_configure_class(cl) -> None
	
	Take a subclass of sprites.Graphic and give it an appropriate render_container."""
	
	upgroups = self._upgroups
	isdialog = (cl.__name__.count("Dialog")) #KLUDGE
	upda = cl.update_auto
	for clgroup in self.clgroups:
	    inup = (clgroup in upgroups)
	    if clgroup.clip is cl.render_clip and clgroup.layer is cl.render_layer\
	    and\
	    (
	    (inup and upda) or (not inup and not upda) #i.e. they match re updating
	    )\
	    and\
	    (
	    (isinstance(clgroup, DialogGroup) and isdialog) or (not isinstance(clgroup, DialogGroup) and not isdialog)
	    ):
		cl.render_container = clgroup
		break
	else:
	    if isdialog:
		newgroup = DialogGroup(cl.render_layer, cl.render_clip)
	    else:
		newgroup = ClipGroup(cl.render_layer, cl.render_clip)
	    cl.render_container = newgroup
	    if cl.update_auto: 
		self._upgroups.append(newgroup)
	    self.clgroups.append(newgroup)


    def add_class(self, cl):#NOT USED
	"""-> None
	
	Configure a single class."""
	
	self._configure_class(cl)    
	self._sort_clgroups()


    def draw_all(self, surface):
	"""draw_all(surface) -> list of rects drawn to
	
	Draw all sprites in all clipgroups, in the right order."""
	
	dirtyrects = []
	for clgroup in self.clgroups: dirtyrects += clgroup.draw(surface)
	return dirtyrects


    def clear_all(self, surface, background):
	"""clear_all(surface, background) -> None
	
	Clear the surface for each clipgroup."""
	
	for clgroup in self.clgroups: clgroup.clear(surface, background)

    
    def update_autos(self, *args, **kwargs):
	"""update_autos(*args, **kwargs) -> None
	
	Call update() on the updating groups."""
	
	for clgroup in self._upgroups: clgroup.update(*args, **kwargs)

    
    def get_sprite_at(self, pos):
	"""get_sprite_at(pos) -> Sprite or None

	Get the top-layer sprite, if any, at coordinate position pos."""

	grps = self.clgroups
	for i in range(len(grps)):
	    for spr in grps[-(i + 1)].sprites():
		if spr.rect.collidepoint(pos): return spr
       
       
    def get_TT_sprite_at(self, pos):
	"""get_TT_sprite_at(pos) -> Sprite or None

	Get the top-layer tool-tip-able sprite, if any, at coordinate position pos.
	
	A tooltip-able sprite is considered to be one with a tt_text attribute
	that is not None."""

	grps = self.clgroups
	for i in range(len(grps)):
	    for spr in grps[-(i + 1)].sprites():
		if spr.tt_text is not None and spr.rect.collidepoint(pos):
                    if spr.render_clip is None or spr.render_clip.collidepoint(pos): #SE from bug 726416
		        return spr 


class ClipGroup(pygame.sprite.RenderUpdates):
    """Draws its sprites to a particular clipping rectangle.
    
    Attributes:
	layer	-- int -- lowest layer is 0
	clip	-- Rect or None -- clipping rectangle
	"""

    draw2 = pygame.sprite.RenderUpdates.draw #quicker this way
    

    def __init__(self, layer, clip, sprite=()):

	#print >> DEBUG_STREAM, \
	#"%s --> init, layer:%s, clip:%s, sprite:%s"%(self.__class__.__name__, layer, clip, sprite) #DEBUG

	self.layer = layer
	self.clip = clip
	
	pygame.sprite.RenderUpdates.__init__(self, sprite)


    def draw(self, surface):

	surface.set_clip(self.clip)
	return self.draw2(surface)


class DialogGroup(ClipGroup):
    """Draws its sprites first in, first drawn."""
    # this is a really crufty class, but it's not used heavily so fine.

    def __init__(self, layer, clip, sprite=()):
	
	self.list = []
	ClipGroup.__init__(self, layer, clip, sprite)

    
    def add_internal(self, sprite):

	#hmm, how best to do this really?
	if sprite not in self.list:
	    self.list.append(sprite)
	ClipGroup.add_internal(self, sprite)

	
    def remove_internal(self, sprite):

	while sprite in self.list:
	    self.list.remove(sprite)
	ClipGroup.remove_internal(self, sprite)


    def draw(self, surface):

	surface.set_clip(self.clip) 
	# Code copying from RenderUpdates
	spritedict = self.spritedict
        surface_blit = surface.blit
        dirty = self.lostsprites
        self.lostsprites = []
        dirty_append = dirty.append
        for s in self.list: #here's the difference
	    # Draw in order!
	    r = spritedict[s]
            newrect = surface_blit(s.image, s.rect)
            if r is 0:
                dirty_append(newrect)
            else:
                dirty_append(newrect.union(r))
            spritedict[s] = newrect
        return dirty
