# Sulk - sprites - anim

#*****************************************************************************
#
#                                   Sulk
#
#       Copyright (c) 2002 Toby Woodwark <zirtix@softhome.net>
#
# This program is free software.  See the file COPYING for details.
#
#*****************************************************************************

"""
Sulk animated graphic classes.
"""

##############################################################################
####### Import

## Modules
import pygame

## My modules
from constants import *  
import graphic


##############################################################################
####### Piece anim

class PieceAnim(graphic.Graphic):
    """Cycles thru self.images.  Up to one may be attached to a Piece instance
    at once.

    If the 'sound' attribute is true, it is played and stopped in sync with the
    visual.

    Parent class.
    
    Attributes:
    (class)
    	_hidepiece  -- bool -- if true, while running the piece must be hidden
	images	    -- list of images
	pattern	    -- tuple of ints -- how long each image lasts in msec, optimally
	iter	    -- int -- number of times to run the pattern
	sound	    -- pygame Sound or None
	sound_fadeout -- int -- milliseconds grace for the sound to die when 
			    kill() is called (if None, allow sound to play on)
    (instance)
	piece	    -- Piece instance
	images	    -- list of images -- rotated to suit
	patternindex -- int -- which image we are using
	facing	    -- int -- piece's initial facing
	iter_remain -- int -- how many anim cycles are left
	msec_remain -- int -- how many msec left for current image
	channel	    -- pygame Channel or None
	"""

    render_layer = 6
    render_clip = MAP_RECT
    update_auto = True

    _hidepiece = False
    iter = 1
    sound = None
    sound_fadeout = None
    channel = None


    def __init__(self, piece):

	graphic.Graphic.__init__(self)
	
	if self._hidepiece: self.piece.hide()
	if self.sound: self.channel = self.sound.play()

	# Get rid of any previous anim
	if piece.anims: piece.anims.sprite.kill()
	piece.anims.add(self)
    
    	self.piece = piece
	self.facing = int(piece.facing)
	
	self.images = map(lambda i: pygame.transform.rotate(i, -90 * piece.facing), self.images)
	
	self.image = self.images[0]
	self.msec_remain = self.pattern[0]
	self.patternindex = 0
	self.iter_remain = self.iter


    def update(self):
	
	if self.iter_remain > 0 and self.facing == self.piece.facing\
	and self.piece.alive():
	    self.rect = self.piece.rect
	    self.msec_remain -= self.elapsed
	    if self.msec_remain <= 0:
		# Switch to next image in self.images
		if self.patternindex == 0:
		    self.iter_remain -= 1
		self.patternindex = (self.patternindex + 1) % len(self.pattern)
		self.msec_remain = self.pattern[self.patternindex]
		self.image = self.images[self.patternindex]
	else:
	    self.kill()

    
    def kill(self):
	
	if self.piece.alive() and self._hidepiece:
	    self.piece.show()
	if self.channel and self.sound_fadeout is not None:
	    self.channel.fadeout(self.sound_fadeout)
	    
	graphic.Graphic.kill(self)


class Flash(PieceAnim):
    """For shooting.

    Anim attributes depend on the piece's weapon.
    
    Attributes:
    (class)
	images_map  -- map of weapon:list of images -- provided externally
	sound_map   -- map of weapon:sound -- provided externally
	pattern_map -- map of weapon:pattern -- how long to show each image
	iter_map    -- map of weapon:iter -- how many times to run the pattern
        sound_fadeout_map -- map of weapon:int milliseconds
	"""
    
    pattern_map = {
    GUN:(80,40), 
    FLAMEGUN:(400,),
    ASSAULT_CANNON:(500,900)}
    iter_map =	{
    GUN:8,
    FLAMEGUN:1,
    ASSAULT_CANNON:2}
    sound_fadeout_map =	{
    GUN:300,
    FLAMEGUN:None,
    ASSAULT_CANNON:None} #hmm

    
    def __init__(self, piece, weapon):
	
	self.images     = self.images_map[weapon]
	self.pattern    = self.pattern_map[weapon]
	self.iter       = self.iter_map[weapon]
	self.sound      = self.sound_map[weapon]
	self.sound_fadeout = self.sound_fadeout_map[weapon]

        PieceAnim.__init__(self, piece)


##############################################################################
####### Stuff that crops up in HUD / MAP_RECT

class SelectDec(graphic.BoardGraphic):
    """The selection indication square.
    
    There is up to one of these at once.
    Has a pulsating image.
    
    Attributes:
    (class)
	image, cat_image
	_min_fade_rate	-- int
	_initial_alpha	-- int
	_min_alpha	-- int
	_max_alpha	-- int
    (instance)
	_waxing		-- -1 or 1
	piece		-- a Piece -- the selected piece
	"""

    render_layer = 5
    render_clip = MAP_RECT
    update_auto = 1

    tt_text = None
      
    _min_fade_rate = 10
    
    _initial_alpha = 150
    _min_alpha = 50
    _max_alpha = 248
    
    
    def __init__(self, piece):
	
        graphic.BoardGraphic.__init__(self)
	self.piece = piece
	if self.piece.type == CAT: self.image = self.cat_image #hmm
        self.image.set_alpha(self._initial_alpha)
        self._waxing = 1
	self.rect = self.image.get_rect()
        self.rect.center = self.piece.rect.center


    def update(self):
	
	if self.piece.alive():
	    self.rect.center = self.piece.rect.center
	    a = self.image.get_alpha()
	    if self._min_alpha > a:
		self._waxing = -self._waxing
		self.image.set_alpha(self._min_alpha)
	    elif self._max_alpha < a:
		self._waxing = -self._waxing
		self.image.set_alpha(self._max_alpha)
	    else:
		self.image.set_alpha(a + int(self._waxing * self.speed_ratio * self._min_fade_rate))
	else:
	    self.kill()


class FacingDec(graphic.BoardGraphic):
    """The facing-choosing arrow.
    
    Has a pulsating image.
    
    Attributes:
    (class)
	_min_fade_rate	-- int -- alpha delta per tick at max frame rate
	_initial_alpha	-- int
	_distratio	-- float -- how much of an SQ_LENGTH these stand away from their piece
    (instance)
	key		-- int -- a pygame key
	piece		-- a Piece
	facing		-- int
	_waxing		-- -1 or 1 -- glow shrinking or growing
	"""

    render_layer = 6
    render_clip = MAP_RECT
    update_auto = 1

    tt_text = None
    
    _distratio = 0.8
    _min_alpha = 100
    _max_alpha = 200

    _min_fade_rate = 5
    
    
    def __init__(self, piece, facing):
	
        graphic.BoardGraphic.__init__(self)
	self.piece = piece
	self.facing = facing
	self.image = pygame.transform.rotate(self.image, -90 * self.facing)
	self.rect = self.image.get_rect()

	self.rect.center = (self.piece.rect.centerx +\
		    int(self._distratio * FACEMAP[self.facing][0]),\
		    self.piece.rect.centery +\
		    int(self._distratio * FACEMAP[self.facing][1]))
	
        self.image.set_alpha(self._max_alpha)
        self._waxing = -1
	self._bounced = 0

	if self.facing == UP: self.key = K__F
	elif self.facing == RIGHT: self.key = K__R
	elif self.facing == DOWN: self.key = K__B
	elif self.facing == LEFT: self.key = K__L


    def update(self):
	
	if not self.piece.alive(): 
	    self.kill()
	    return
	if self._bounced >= 2: return
        a = self.image.get_alpha()
	if a < self._min_alpha:
	    self._waxing = -self._waxing
	    self.image.set_alpha(self._min_alpha)
	elif a > self._max_alpha:
	    self._waxing = -self._waxing
	    self.image.set_alpha(self._max_alpha)
	    self._bounced += 1
	else:
	    self.image.set_alpha(a + int(self._waxing * self.speed_ratio * self._min_fade_rate))
