# Sulk - sprites - buttons 

#*****************************************************************************
#
#                                   Sulk
#
#       Copyright (c) 2002 Toby Woodwark <zirtix@softhome.net>
#
# This program is free software.  See the file COPYING for details.
#
#*****************************************************************************

"""
Sulk buttons.
"""

##############################################################################
####### Import / set up

## My modules
from constants import * 
import util
import functions
import graphic
import misc
import textual
#N.B. cannot import 'players' into this module


###############################################################################
####### Buttons

class Button(graphic.Graphic):
    """Pressable UI element.  
    
    Parent class.  
    
    Attributes:
    (class)
	key		-- int -- a Pygame key
	_default_status -- int
	all_buttons	-- spritegroup
	img		-- dict of status:surface
	unpress_sound	-- sound -- played when successfully unpressed
	press_sound	-- sound -- played when successfully pressed
	fail_sound	-- sound -- played when unsuccessfully  hit
	tt_text_dict	-- dict of status:string -- for tt_text
    (instance)
	status		-- int
    (optional)
	spawned		-- an InputHandler (see UIinput classes)
	"""
	
    render_layer = 0
    render_clip = None
    update_auto = False

    key = None
    
    tt_text = lambda self: self.tt_text_dict[self.status]
    
    def __init__(self):
	
        graphic.Graphic.__init__(self)
	self.add(self.all_buttons)
	self.status = self._default_status
        self.update()
        self.rect = self.image.get_rect()


    def update(self):
	
	self.image = self.img[self.status]


    def press(self):
	
	self.status = PRESSED
	self.update()#?


    def unpress(self):
	
	self.status = ENABLED	
	self.update()


    def hit(self):
	"""hit() -> None
	
	Toggle status if possible."""
	
	if self.status == ENABLED: 
	    self.press()
	    self.press_sound.play()
	elif self.status == PRESSED: 
	    self.unpress()
	    self.unpress_sound.play()	    
	else:
	    self.fail_sound.play()


##############################################################################
####### 'Generic' buttons

class GlobalButton(Button):
    """Attaches to a SulkGame instance.
    """
    
    update_auto = True #?
    _default_status = DISABLED

    def __init__(self, sulkgame):
	
	self.game = sulkgame
	Button.__init__(self)


class CPButton(GlobalButton):
    """If pressed, Marines use CPs, not APs.
    """
    
    tt_text_dict = {
    ENABLED:"Use command points",
    DISABLED:"Toggle command points",
    PRESSED:"Stop using command points"}
    
    key = K__CP
    

    def __init__(self, sulkgame):
	
        GlobalButton.__init__(self, sulkgame)	
        self.rect.topleft = EXTRABUTTONS_RECT.topleft


    def press(self):
	
        GlobalButton.press(self)
	self.game.phase.cpiece().buttons.update() #using Nullpower


    def unpress(self):
	
	GlobalButton.unpress(self)
	self.game.phase.cpiece().buttons.update() #using Nullpower


class PauseButton(GlobalButton):
    """Brings up main menu.
    """
    
    tt_text = lambda self: "Pause the game"
    
    key = K__PAUSE_GAME
    

    def __init__(self, sulkgame):
	
        GlobalButton.__init__(self, sulkgame)
        self.rect.topleft = EXTRABUTTONS_RECT.topleft	
	self.rect.move_ip(
	int(self.image.get_rect().width / 2),
	int(self.image.get_rect().height * 0.90) #squish up next to upper buttons
	) #kludgey


    def update(self):

	if self.game.paused:
	    self.status = PRESSED
	elif getattr(getattr(self.game, 'phase', None), 'pausable', None):
	    self.status = ENABLED
	else:
	    self.status = DISABLED
	GlobalButton.update(self)
	
    
    def press(self):
    
	ih = self.game.marines.player.UIinput.Handle_Main_Menu(self.game.phase) #FIXME player ref
	ih.controlling_button = self #?
	GlobalButton.press(self)


class SecretsButton(GlobalButton):
    """Toggles team secrets.
    """
    
    tt_text_dict = {
    ENABLED:"Show secrets",
    DISABLED:"Toggle secrets",
    PRESSED:"Hide secrets"}
    
    key = K__SECRET
    

    def __init__(self, sulkgame):
	
	GlobalButton.__init__(self, sulkgame)
        self.rect.topleft = EXTRABUTTONS_RECT.topleft	
	self.rect.move_ip(
	int(3 * self.image.get_rect().width / 2),
	int(self.image.get_rect().height * 0.90) #squish up next to upper buttons
	) #kludgey


    def update(self):

	try:
	    team = self.game.phase.secrets_team
	except:
	    self.status = DISABLED
	else:
	    if team == None:
		self.status = DISABLED #KLUDGE
	    else:
		if team.secrets_showing:
		    self.status = PRESSED
		else:
		    self.status = ENABLED
	GlobalButton.update(self)
	
  	
    def unpress(self):
    
	self.game.phase.secrets_team.hide_secrets()
	GlobalButton.unpress(self)


    def press(self):
    
	self.game.phase.secrets_team.show_secrets()
	GlobalButton.press(self)


class DoneButton(GlobalButton):
    """Ends turns and interrupts.
    
    See Phase subclasses.
    """
    
    key = K__DONE

    def tt_text(self):
	if getattr(self.game.phase, 'isinterrupt', 0):
	    return "Cancel this interrupt"
	return "End this action phase"

    
    def __init__(self, sulkgame):
	
        GlobalButton.__init__(self, sulkgame)
        self.rect.topleft = EXTRABUTTONS_RECT.topleft	
	self.rect.move_ip(self.image.get_rect().width, 0) #kludge
	

##############################################################################
####### Piece buttons

class PieceButton(Button):
    """Belongs to a piece.
    
    We only update() PieceButtons when a Piece moves.  NOT every tick (see
    SulkGame loop)

    Extra attributes:
	piece -- Piece instance

    Some subclassed buttons also have the attribute:
	spawned -- an InputHandler of some kind
    and a method defined outside this module:
	spawn_magic() -- spawn and return an InputHandler instance
    """
    # FIXME RULES these updates are a breach of hotseat secrecy! (how about a
    # config option?)
    
    _default_status = ENABLED
    
    
    def __init__(self, piece):
	
	self.piece = piece
	Button.__init__(self)
        self.rect.left = PIECEBUTTONS_RECT.left
	#move the button down to first available slot in PIECEBUTTONS_RECT
        self.rect.top = PIECEBUTTONS_RECT.top + (len(self.piece.buttons.sprites()) * BUTTON_SIZE[1]) 
	self.add(self.piece.buttons)
	self.update()


    def press(self):
	
	#placeholder
	Button.press(self)
	self.update()


    def unpress(self):
	
	#placeholder
	Button.unpress(self)
	self.update()


    def update(self):
	
	if self.piece.alive(): Button.update(self)
	else: self.kill()


    def _disable_all_others(self): #DOCSTRING?

	for b in self.piece.buttons.sprites():
	    if b is not self:
		b.status = DISABLED
		#PieceButton.update(b) #no, because we want this to be invisible


    def _undisable_all_others(self):
	
	self.piece.buttons.update()
	

class AttackButton(PieceButton):
    """Attack in close combat!"""
    #never PRESSED
    
    tt_text = lambda self: "Attack in close combat"

    key = K__ATTACKBUTTON
    
    
    def update(self):
	
	if self.piece.get_points_avail() < self.piece._cc_cost: 
	    self.status = DISABLED
	else:
	    for sq in self.piece.square.get_adjacents():
		if functions.ang(self.piece.square, sq, self.piece.facing) == F:
		    targetsq = sq
		    break
	    else: 
		self.status = DISABLED
		PieceButton.update(self)
		return
	
	    for piece in targetsq.piece.sprites():
		if self.piece.can_cc(piece) is not None:
		    self.status = ENABLED
		    break
	    else:
		for feature in targetsq.features.sprites():
		    if self.piece.can_cc(feature) is not None:
			self.status = ENABLED
			break
		else:
		    self.status = DISABLED
		    
	PieceButton.update(self)


    def press(self):
	
	if self.piece.get_points_avail() < self.piece._cc_cost: return
	# find square
	for sq in self.piece.square.get_adjacents():
	    if functions.ang(self.piece.square, sq, self.piece.facing) == F:
		targetsq = sq
		break
	else: return
	
	#find target
        for piece in targetsq.piece.sprites():
            if self.piece.can_cc(piece) is not None:
                target = piece
                break
        else:
            for feature in targetsq.features.sprites():
                if self.piece.can_cc(feature) is not None:
                    target = feature
                    break
	    else:
		textual.ScrollMessage("Cannot attack.") #hmm?
		return
	self.piece.goCC(target)


class OverwatchButton(PieceButton):
    """Overwatch on/off!
    """
    
    tt_text_dict = {
    DISABLED:"Toggle overwatch",
    ENABLED:"Enable overwatch",
    PRESSED:"Cancel overwatch"}
    
    key = K__OVERWATCHBUTTON


    def update(self):
	
	if self.piece.overwatching: self.status = PRESSED
	elif self.piece.get_points_avail() >= self.piece._ow_cost: 
	    self.status = ENABLED
	else: 
	    self.status = DISABLED
	PieceButton.update(self)


    def press(self):
	
	self.piece.take_points(self.piece._ow_cost)
	PieceButton.press(self)
	self.piece.overwatchon()


    def unpress(self):
	
	self.piece.overwatchoff()
	PieceButton.unpress(self)
    
    
class ShootButton(PieceButton):
    """Shoot!"""
    
    tt_text_dict = {
    DISABLED:"Toggle shoot mode",
    ENABLED:"Enable shoot mode",
    PRESSED:"Cancel shoot mode"}
    
    key = K__SHOOTBUTTON


    def update(self):
	
	if self.status != PRESSED:
	    if not (self.piece.gunjammed or self.piece.overwatching)\
	    and self.piece.ammoremaining\
	    and (
		self.piece.get_points_avail() >= self.piece._sh_cost\
		or self.piece.try_mns
		): 
		self.status = ENABLED
	    else: 
		self.status = DISABLED
	PieceButton.update(self)


    def unpress(self):
	
	PieceButton.unpress(self)
	self.spawned.destroy() #hmm
	del self.spawned
        #for b in self.piece.buttons.sprites():
        #    if isinstance(b, MovenshootButton):
        #        b.unpress() #??????????


    def press(self):
	
        self.spawned = self.piece.team.player.UIinput.Shoot_Something(self.piece.team.game.phase, self.piece)
	self.spawned.controlling_button = self #?
	PieceButton.press(self)

        
class FireGrenadeButton(PieceButton):
    """Fire grenade"""
    
    tt_text_dict = {
    DISABLED:"Toggle fire grenade mode",
    ENABLED:"Enable fire grenade mode",
    PRESSED:"Cancel fire grenade mode"}
    
    #key = K__SPECIAL_MOVE


    def update(self):
	
	if self.status != PRESSED:
	    if not self.piece.overwatching\
	    and self.piece.ammoremaining\
	    and (
		self.piece.get_points_avail() >= self.piece._gr_cost\
		or self.piece.try_mns
		): 
		self.status = ENABLED
	    else: 
		self.status = DISABLED
	PieceButton.update(self)


    def unpress(self):
	
	PieceButton.unpress(self)
	self.spawned.destroy() #hmm
	del self.spawned
        #for b in self.piece.buttons.sprites():
        #    if isinstance(b, MovenshootButton):
        #        b.unpress() #??????????


    def press(self):
	self.spawned = self.piece.team.player.UIinput.Fire_Grenade(self.piece.team.game.phase, self.piece)
	self.spawned.controlling_button = self #?
	PieceButton.press(self)


class UnjamButton(PieceButton):
    """Unjam!
    
    Present iff piece has a jam, i.e. instead of the shoot, 
    overwatch and movenshoot buttons
    """
    #never PRESSED
    
    tt_text = lambda self: "Unjam the weapon"
    
    key = K__SPECIAL_MOVE


    def update(self):
	
	if self.piece.get_points_avail() >= self.piece._unjam_cost: 
	    self.status = ENABLED
	else: 
	    self.status = DISABLED
	PieceButton.update(self)


    def press(self):
	
	self.piece.goUngunjam()

        
class ReloadButton(PieceButton):
    """Reload heavy weapons.
    
    """
    #never PRESSED
    
    tt_text = lambda self: "Reload the weapon"
    
    #key = K__SPECIAL_MOVE


    def update(self):
	
	if self.piece.get_points_avail() >= self.piece._reload_cost and self.piece.reloadsremaining: 
	    self.status = ENABLED
	else: 
	    self.status = DISABLED
	PieceButton.update(self)


    def press(self):
	
	self.piece.goReload()

        
class AutofireButton(PieceButton):
    """Autofire assault cannon.
    """
    #never PRESSED
    
    tt_text = lambda self: "Auto fire!!!!"
    
    key = K__SPECIAL_MOVE


    def update(self):
	
        if self.piece.can_autofire() and self.piece.get_points_avail() >= self.piece._autofire_cost: 
	    self.status = ENABLED
	else: 
	    self.status = DISABLED
        PieceButton.update(self)


    def press(self):
	
	self.piece.goAutofire()

        
class SelfdestructButton(PieceButton):
    """Foom!  Self-destruct.
    """
    
    tt_text_dict = {
    PRESSED:"Cancel self-destruct",
    ENABLED:"Self-destruct",
    DISABLED:"Self-destruct"}
    
    key = K__SPECIAL_MOVE


    def update(self):
	
	if self.status != PRESSED:
	    if self.piece.can_selfdestruct() and self.piece.get_points_avail() >= self.piece._selfdestruct_cost: 
		self.status = ENABLED
	    else: 
		self.status = DISABLED
	PieceButton.update(self)


    def unpress(self):
	
	# Cancel the dialog
	self.tt_text = self.__class__.tt_text
	PieceButton.unpress(self)
	self.spawned.destroy() #hmm
	del self.spawned
	self._undisable_all_others()
	textual.ScrollMessage("Self-destruct cancelled.")


    def press(self):
	
	PieceButton.press(self)
	self.spawned = self.piece.team.player.UIinput.Selfdestruct_A_Piece(self.piece.team.game.phase, self.piece)
	self.spawned.controlling_button = self
	# Disable the other buttons!
	self._disable_all_others()


class UsedoorButton(PieceButton):
    """Use door!
    """
    
    tt_text_dict = {
    PRESSED:"Cancel using a door",
    ENABLED:"Use a door",
    DISABLED:"Use a door"}
    
    key = K__USEDOORBUTTON


    def update(self):
	
	if self.status == PRESSED: 
	    PieceButton.update(self)
	    return
	elif self.piece.get_points_avail() >= self.piece._usedoor_cost:
	    can_use_door = self.piece.can_use_door 
	    Door = misc.Door 
	    for s in self.piece.square.get_adjacents():
		for d in s.features.sprites():
		    if isinstance(d, Door) and can_use_door(d) is not None:
			self.status = ENABLED
			PieceButton.update(self)
			return
	self.status = DISABLED
	PieceButton.update(self)


    def press(self):
	
	# Find nearby doors
	candidates = pygame.sprite.Group()
	can_use_door = self.piece.can_use_door
	Door = misc.Door
	for s in self.piece.square.get_adjacents():
	    for d in s.features.sprites():
		if isinstance(d, Door) and can_use_door(d) is not None:
		    candidates.add(d)
	if candidates:
	    # We have some candidate doors
	    if len(candidates) > 1:
		PieceButton.press(self)
		self._disable_all_others()
		self.spawned = self.piece.team.player.UIinput.Use_A_Door(self.piece.team.game.phase, self.piece, candidates)
		self.spawned.controlling_button = self
	    else: 
		# We have unambiguous choice of door
		self.piece.goDoor(candidates.sprites()[0], hostility=0)


    def unpress(self):
	
	if hasattr(self, 'spawned'): 
	    self.spawned.destroy() #hmm
	    #del self.spawned #hmm, unneeded now I think
	    #textual.ScrollMessage("Door use cancelled.")
	self._undisable_all_others()
	PieceButton.unpress(self)

        
class CutDoorButton(PieceButton):
    """Cut through a door or bulkhead.
    """

    #never PRESSED
    
    tt_text = lambda self: "Cut through a door"
    
    #key = K__USEDOORBUTTON


    def update(self):
	
	if self.status == PRESSED: 
	    PieceButton.update(self)
	    return
	elif self.piece.get_points_avail() >= self.piece._usedoor_cost:
	    can_cut_door = self.piece.can_cut_door 
	    Door = misc.Door 
	    for s in self.piece.square.get_adjacents():
		for d in s.features.sprites():
		    if isinstance(d, Door) and can_cut_door(d) is not None:
			self.status = ENABLED
			PieceButton.update(self)
			return
	self.status = DISABLED
	PieceButton.update(self)


    def press(self):
	
	# Find a cuttable door
        door = None
	for s in self.piece.square.get_adjacents():
	    for d in s.features.sprites():
		if isinstance(d, misc.Door) and self.piece.can_cut_door(d) is not None:
		    door = d
                    break
        if door:
            self.piece.cut_door(door)


class ConvertButton(PieceButton):
    """Deploy stealers!
    """ #blips only
    #never PRESSED
    
    tt_text = lambda self: "Convert to Stealers"
    
    key = K__SPECIAL_MOVE


    def update(self):
	
	if self.status != PRESSED:
	    if self.piece.can_convert(): self.status = ENABLED
	    else: self.status = DISABLED
	PieceButton.update(self)


    def press(self):

	self.piece.convert() 
	#I am now dead
	#no need for PieceButton.press(self)


class MovenshootButton(PieceButton):
    """Move'n'shoot!
    
    If this button is PRESSED, piece does not spend APs to shoot after a move.
    """
    
    tt_text_dict = {
    PRESSED:"Cancel move-and-shoot mode",
    ENABLED:"Enable move-and-shoot mode",
    DISABLED:"Toggle move-and-shoot mode"}
    
    key = K__MOVENSHOOTBUTTON


    def update(self):
	
	# Tricky...
	if self.piece.overwatching: self.status = DISABLED
	elif self.piece.try_mns: self.status = PRESSED
        else: self.status = ENABLED
#	elif self.piece.get_points_avail(): self.status = ENABLED
#	else: self.status = DISABLED
	PieceButton.update(self)


    def press(self):
	
#        for b in self.piece.buttons.sprites():
#            if isinstance(b, ShootButton):
#                b.press()
        self.piece.try_mns = True
	self.update()


    def unpress(self):
	
	self.piece.try_mns = False
	self.update()


class CATButton(PieceButton):
    """Use a C.A.T. that piece shares its square with.

    Button is PRESSED just when the piece is carrying a C.A.T.,
    ENABLED if the C.A.T. is at the piece's feet, and DISABLED if 
    otherwise (but we are in a C.A.T.-using mission).
    
    If we are hit when PRESSED, behaviour is to attempt to pass the C.A.T..
    If we are hit when ENABLED, try to pick up the C.A.T..

    Attributes:
	cat	-- a C_A_T instance or None
	spawned -- a Pass_CAT instance
    """

    key = K__CATBUTTON

    tt_text_dict = {
    PRESSED:"Pass C.A.T. to another Marine",
    ENABLED:"Pick up the C.A.T.",
    DISABLED:"C.A.T. actions"}
    
    #TODO -- multiple CATs per marine ?!

    def update(self):

	if self.piece.get_points_avail() and not self.piece.overwatching:
	    # Are we holding the C.A.T.?
	    for x in self.piece.possessions.sprites():
		if x.type == CAT: #kludgey
		    self.cat = x
		    self.status = PRESSED
		    break
	    else:
		# Is the C.A.T. available?
		for x in self.piece.square.features.sprites():
		    if x.type == CAT:
			self.cat = x
			self.status = ENABLED
			break
		else:
		    self.cat = None
		    self.status = DISABLED
	else:
	    self.cat = None
	    self.status = DISABLED
	PieceButton.update(self)


    def press(self):
	
	self.piece.goPickup_cat(self.cat)
	self.update()


    def unpress(self):

	if hasattr(self, 'spawned'): 
	    self.spawned.destroy()
	    self._undisable_all_others()
	    #del self.spawned #done by destroy
	else: 
	    candidates = pygame.sprite.Group()
	    can_pass_cat = self.piece.can_pass_cat
	    cat = self.cat
	    for p in self.piece.team.sprites():
		if can_pass_cat(cat, p):
		    candidates.add(p)
	    if not candidates: 
		textual.ScrollMessage("Nowhere to pass the C.A.T..")
		return
	    elif len(candidates) == 1:
		if not self.piece.goPass_cat(cat, candidates.sprites()[0]):
		    textual.ScrollMessage("Cannot pass the C.A.T..")		    
		return
	    else:
		self._disable_all_others()
		# Make an inputhandler for passing the cat
		self.spawned = self.piece.team.player.UIinput.Pass_CAT(self.piece.team.game.phase, self.piece, self.cat, candidates)
		self.spawned.controlling_button = self #?
	self.update()
