# Sulk - sprites - functions

#*****************************************************************************
#
#                                   Sulk
#
#       Copyright (c) 2002 Toby Woodwark <zirtix@softhome.net>
#
# This program is free software.  See the file COPYING for details.
#
#*****************************************************************************

"""Sulk graphic functions."""

#TODO (perhaps) -- replace los-checking and angle algorithms with methods of
# a global object - that knows/caches what heading any square is
# from any given square.  (possibly use board.Board)
# (or) use Square methods and cache this info in Square attributes (a la
# adjacents). - TW

##############################################################################
####### Import

## Modules
import pygame, math

## My modules
from constants import *
import misc # for Square

    	   
##############################################################################	   
####### Functions

## Movement
def find_shortest_routes(square1, square2, square_test=lambda sq: True, maxlen=None):
    """find_shortest_routes(square1, square2, square_test, maxlen) -> list of lists
    find shortest routes from square1 to square2, ignoring obstacles

    Returns a possibly empty list of routes.  
    Each route a list of squares to traverse in order: [square1, ..., square2].  
    These squares will be orthogonally or diagonally adjacent in series.
    
    If given, applies the unary function 'square_test' to each square on the route 
    (apart from the endpoints) and only uses those for which it is true.
    
    If given, restricts routes to a length of 'maxlen', including endpoints.  If no
    routes so short are found, will return []."""
    
    # Depends on all squares being connected
    # IN PROGRESS -- needs to be replaced with A*
    
    if square1 is square2: return [square1]
    shortestroutes = []
    routes = [[square1]]
    visited_dict = {square1.coord:True} #used to store coords of squares we have visited so far
    
    while True:
	#print >> DEBUG_STREAM,\
        #"find_shortest_routes ::: while 1" #DEBUG
	#print >> DEBUG_STREAM,\
        #"find_shortest_routes ::: while 1 ::: number of current routes:",len(routes) #DEBUG
        # Give up when we have exhausted our search tree
	if not routes or (maxlen and len(routes[0]) > maxlen):
            print >> DEBUG_STREAM,\
            "find_shortest_routes, returning no routes" #DEBUG
            return []
	#print >> DEBUG_STREAM,\
        #"find_shortest_routes ::: while 1 ::: length of routes:",len(routes[0]) #DEBUG
	newroutes = []
        append = newroutes.append
	for r in routes:
	    # Find squares reachable from end of route
	    for nextsq in r[-1].get_adjacents():
		if nextsq is square2:
		    # We have got a shortest route
		    shortestroutes.append(r + [square2])
		    break
                # The square must pass the square_test
                if square_test(nextsq):
                    # The next square cannot be in any existing route (?)
                    if visited_dict.get(nextsq.coord):
                        continue
                    else:
                        append(r + [nextsq])
                        visited_dict[nextsq.coord] = True
		    
	if shortestroutes: 
	    print >> DEBUG_STREAM,\
            "find_shortest_routes, returning %d routes of length %d"%\
            (len(shortestroutes),len(shortestroutes[0])) #DEBUG
	    return shortestroutes
        routes = newroutes

    
## Arcs
def ang(s1, s2, facing=UP):
    """ang(s1, s2, facing=UP) -> int or None
    get a heading between any 2 sprites""" 
    
    return heading(s1.rect.center, s2.rect.center, facing)


def heading(pos1, pos2, facing=UP):
    """heading(pos1, pos2, facing=UP) -> int or None

    Takes 2x coordinate tuples (x,y).
    Returns one of F, F_R_F, F_R, F_R_R, R, B_R_R, B_R, B_R_B, 
    B, B_L_B, B_L, B_L_L, L, F_L_L, F_L, F_L_F; or None
    for the heading of pos2 from pos1.  
    Facing is taken from pos1.  So if it is UP we have F = North, R = East, etc"""
    
    if pos1 == pos2: return
    rdist, fdist = pos2[0] - pos1[0], pos1[1] - pos2[1]
    # Adjust results
    for x in range(facing): 
    	fdist, rdist = rdist, -fdist #'rotate map right'
    # Now see what to return
    if fdist == 0:  # to the side
    		if rdist > 0: return R
    		else: return L
    elif fdist > 0: # ahead somewhere
    	if rdist == 0: return F
    	elif rdist > 0: #look right
    		if rdist == fdist: return F_R
    		elif rdist > fdist: return F_R_R
    		else: return F_R_F
    	else:	#look left
    		if -rdist == fdist: return F_L
    		elif -rdist < fdist: return F_L_F
    		else: return F_L_L
    else:	# behind somewhere
    	if rdist == 0: return B
    	elif rdist > 0: #look right
    		if rdist == -fdist: return B_R
    		elif rdist > -fdist: return B_R_R
    		else: return B_R_B
	else:	#look left.
		# we know both rdist and fdist are negative
    		if rdist == fdist: return B_L
    		elif rdist < fdist: return B_L_L #changed for 0.24 - TW
    		else: return B_L_B


## Shooting, overwatch
def find_range(thing1, thing2):
    """find_range(thing1, thing2) -> int
    
    Finds the number of squarelengths between one thing with a .rect and
    another as the crow flies, rounded up."""
    
    xdist = thing1.rect.centerx - thing2.rect.centerx
    ydist = thing1.rect.centery - thing2.rect.centery
    
    # Pythagoras
    dist = math.sqrt(xdist ** 2 + ydist ** 2)
    
    return int(math.ceil(dist / SQ_LENGTH))


def discover_blips(piecelist):
    """discover_blips(piecelist) -> None
    find and convert blips
    
    Checks if any piece in the list has discovered any blips and forces
    conversion.  For our purposes blips are those things with a 'convert'
    method."""
    
    seen = pygame.sprite.Group()
    for piece in piecelist:
	if piece.alive():
	    for spr in piece.team.enemy_team.sprites():
		if hasattr(spr, 'convert') and piece.can_see(spr.square): 
                    seen.add(spr)
    # Seen blips are involuntarily converted
    for spr in seen.sprites(): spr.convert(spr.team.enemy_team)
    # Tidy refs
    seen.empty()
		

## LOS
def _line_seg_intersect(line1point1, line1point2, line2point1, line2point2):
    """_line_seg_intersect(line1point1, line1point2, line2point1, line2point2) -> tuple or int
    
    Returns a point (2-tuple) if the two lines cross, DONT_INTERSECT if not, or COLINEAR if they cross everywhere."""
    
    #This code from David Clark
    
    def have_same_signs(a, b):
	"""have_same_signs(a, b) -> bool

	Returns true if args have the same sign."""
	return ((a ^ b) >= 0)

    x1 = line1point1[0]
    y1 = line1point1[1]
    x2 = line1point2[0]
    y2 = line1point2[1]
    x3 = line2point1[0]
    y3 = line2point1[1]
    x4 = line2point2[0]
    y4 = line2point2[1]

    a1 = y2 - y1  
    b1 = x1 - x2  
    c1 = (x2 * y1) - (x1 * y2)

    r3 = (a1 * x3) + (b1 * y3) + c1  
    r4 = (a1 * x4) + (b1 * y4) + c1

    if ((r3 != 0) and (r4 != 0) and have_same_signs(r3, r4)):
        return(DONT_INTERSECT)

    a2 = y4 - y3  
    b2 = x3 - x4  
    c2 = x4 * y3 - x3 * y4

    r1 = a2 * x1 + b2 * y1 + c2  
    r2 = a2 * x2 + b2 * y2 + c2

    if ((r1 != 0) and (r2 != 0) and have_same_signs(r1, r2)):  
         return(DONT_INTERSECT)

    denom = (a1 * b2) - (a2 * b1)  
    if denom == 0:  
        return(COLINEAR)
    elif denom < 0:
        offset = (-1 * denom / 2)
    else:
        offset = denom / 2
    
    num = (b1 * c2) - (b2 * c1)
    if num < 0:
        x = (num - offset) / denom
    else:
        x = (num + offset) / denom

    num = (a2 * c1) - (a1 * c2)  
    if num <0:
        y = (num - offset) / denom
    else:
        y = (num - offset) / denom

    return (x, y)
    # End of David Clark code
    
    
def get_square_at(pos):
    """get_square_at(pos) -> Square instance / None
    
    Returns the Square instance at position pos, if any."""
    
    for sq in misc.Square.all_squares.sprites(): #cruft
	if sq.rect.collidepoint(pos): 
	    return sq
	    
	    
def seeable(s1, s2):
    #hmm should this be a method of square?
    # TODO cache result in both squares' attributes somehow (e.g. in a dict)
    # in certain cases (i.e. when LOS will *never* be possible)
    """seeable(s1, s2) -> bool
    
    Returns 1 if there is a Space Hulk LOS line between the centres of squares
    s1 and s2.  This does not imply anything about the LOS *arc*.  This funct
    makes certain assumptions about things lining up."""
    
    def poss_seeable(c1, c2, numpoints=4):
	"""poss_seeable(c1, c2, numpoints=4) -> bool
	see if c1 can *possibly* see c2.
    
	Designed to speed up los checks over long distances at funny angles.
	c1 and c2 are the square centres.  Takes numpoints points along the supposed
	los line.  If any are not in a square, return 0.  This is only used in
	the 'gnarly' section of seeable."""
    
	# hmm, Should we check only 2 or 3 points for better speed?
	# or more points for better accuracy?
	numpoints = int(numpoints)
	frac = (float(1) / (numpoints + 1))
	x1, y1, x2, y2 = c1[0], c1[1], c2[0], c2[1]
	offsetx = frac * (x2 - x1)
	offsety = frac * (y2 - y1)
	x, y = x1, y1
	while numpoints:
	    numpoints -= 1
	    x = int(x + offsetx)
	    y = int(y + offsety)
	    if not get_square_at((x,y)): return False
	return True
	
    # Go
    # Check LOS makes sense (i.e. we have Squares and not a LimboSquares)
    if not (s1.section and s2.section): return False
    
    # Get square centers
    c1, c2 = s1.rect.center, s2.rect.center
    if c1 == c2:
	# The squares are the same
	return True
    a = ang(s1, s2) 
    #print >> DEBUG_STREAM, 'seeable ::: ang:', a #DEBUG
    
    # Check the angle
    if a in (F, B, L, R):
	# Orthogonal case ; cool.
	xy = c1
	while True:
	    xy = (xy[0] + RIGHTMAP[a], xy[1] - FORWARDMAP[a])
	    if xy == c2: return True
	    sq = get_square_at(xy)
	    if not sq or not sq.losthru(): return False
	    
    elif a in (F_R, F_L, B_R, B_L): 
	# Exactly diagonal; cool-ish.
	xy = c1
	sq = nextsq = s1
	while True:
	    xy = (xy[0] + RIGHTMAP[a], xy[1] - FORWARDMAP[a])
	    nextsq = get_square_at(xy)
	    if not nextsq: return False
	    # Special check for diagonal los starts here
	    seesqs = []
	    for s in sq.get_adjacents():
		if s in nextsq.get_adjacents() and s.losthru(): seesqs.append(s)
	    if not seesqs: return False
	    if nextsq is s2: return True
	    if not nextsq.losthru(): return False
	    sq = nextsq
	    
    else: 
	# Gnarly inbetween-angle business
	#TODO really work this bitch out and optimize
	#print >> DEBUG_STREAM, 'seeable ::: gnarly, begins' #DEBUG

	# Try to avoid computing this
	if not poss_seeable(c1, c2):
	    #print >> DEBUG_STREAM, 'seeable ::: gnarly, defused!' #DEBUG
	    return 0
	
	if a in	  (F_L_L, F_L_F): dir = F_L # los line goes southeast->northwest
	elif a in (F_R_R, F_R_F): dir = F_R # los line goes southwest->northeast
	elif a in (B_R_R, B_R_B): dir = B_R # los line goes northwest->southeast
	else: dir = B_L # los line goes northeast->southwest
        
	#print >> DEBUG_STREAM, "seeable ::: gnarly, dir:", ANGLENAMES[dir] #DEBUG

	# Loop through squares adjacent to sq looking for a line of squares to s2.
	# sq is the main variable in a tree search.
	xy = c1
	sq = s1
	
	# Keep track of previously-visited 'sq' squares and prevent revisiting.
	visited = []

	while True:
	    #print >> DEBUG_STREAM, "seeable ::: gnarly ::: while 1, begins"  #DEBUG
            
	    visited.append(sq)
	    orthog = []
	    diag = []
	    for s in sq.get_adjacents():
		if s not in visited:
		    if ang(sq, s) in (F, R, B, L): orthog.append(s)
		    else: diag.append(s)
		
	    #print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1, orthogs:', len(orthog) #DEBUG
	    #print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1, diags:', len(diag) #DEBUG
	    
	    for s in orthog:
		#print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in orthog, begins' #DEBUG
	        if s is s2: 
		    # Done -- this is how we break the while
		    return True
		elif s.losthru(): #s is a candidate for new sq:
		    #print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in orthog, LOSable' #DEBUG
		    if dir in (F_L, B_R):
			# We take northeast/southwest lines between the nearby
			# squares' corners to find the square in the direction
			# of s2.
		        newxy = _line_seg_intersect(xy, c2, s.rect.bottomleft, s.rect.topright) #COLINEAR impossible here
		    else:
		        # We take northwest/southeast lines.
			newxy = _line_seg_intersect(xy, c2, s.rect.bottomright, s.rect.topleft) #COLINEAR impossible here
		    if newxy: #i.e. if we found that s was in the right direction:
		        sq = s
			xy = newxy
			#print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in orthog, s is new sq, break back to while' #DEBUG
			break #back to while
	    
	    else:
		#print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1, all orthogs exhausted' #DEBUG
		# We only continue if the los line does not exit sq orthogonally
		for s in diag: 
		    # N.B. Here, we may only take follow los between 2 adjacent squares
		    # diagonally if the los line travels exactly across a corner
		    #print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in diag, begins' #DEBUG
		    if s.losthru():# WTF???--> and ang(sq, s) == a: #s is candidate for new sq
			
			#print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in diag, LOSable' #DEBUG
			#N.B. s can never be a candidate for s2 in this loop. (don't ask)
		        if dir in (F_L, B_R):
			    # We take northeast/southwest lines between the nearby squares' corners to find the one
		            # in the direction of s2.
			    newxy = _line_seg_intersect(xy, c2, s.rect.bottomleft, s.rect.topright) #COLINEAR impossible here
		        else:
			    # We take northwest/southeast lines.
			    newxy = _line_seg_intersect(xy, c2, s.rect.bottomright, s.rect.topleft) #COLINEAR impossible here
			#print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in diag, newxy:',newxy #DEBUG
			#print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in diag, xy:',xy #DEBUG
			if newxy:
			    dist_a = abs(xy[0] - sq.rect.centerx) #'old'
			    dist_b = abs(newxy[0] - s.rect.centerx)#'new'
			    #print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in diag, dist_a,dist_b:',dist_a,dist_b #DEBUG
			    if dist_a == dist_b: #(i.e. s really is the 'next' square(maybe this should be within +-1?)):
				# Find a LOSable common neighbour of sq and s
				for adj_sqr in sq.get_adjacents():
				    if adj_sqr in s.get_adjacents() and adj_sqr.losthru(): 
					break #back to for
				else: 
				    #print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in diag, no LOSable common neighbour' #DEBUG
				    #No such neighbour, we fail
				    return False
				# Now we can go on to next
				sq = s
				xy = newxy
				#print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: s in diag, s is new sq, break back to while' #DEBUG
				break #back to while
		
		else: 
		    # No 'break' so we've run out of squares in both the 'for' loops
		    #print >> DEBUG_STREAM, 'seeable ::: gnarly ::: while 1 ::: diag exhausted; returning 0' #DEBUG
		    return False
