# Sulk - sprites - misc

#*****************************************************************************
#
#                                   Sulk
#
#       Copyright (c) 2002 Toby Woodwark <zirtix@softhome.net>
#
# This program is free software.  See the file COPYING for details.
#
#*****************************************************************************

"""
Sulk miscellaneous graphic classes.
"""

##############################################################################
####### Import / set up

## Modules
import pygame
import random

## My modules
import util
from constants import *  
import functions
import graphic
import textual


##############################################################################
####### Effects

class HiLite(graphic.Graphic):
    """Highlights squares:
	
	GREENHILITE square can be deployed to
	REDHILITE square has an object that can be used/chosen etc.
    
    Attributes:
    (class)
	img	    -- dict of int:surface
	"""
	
    render_layer = 6
    render_clip = MAP_RECT
    update_auto = 0

    
    def __init__(self, square, colortype=GREENHILITE):
	
        graphic.Graphic.__init__(self, square.hilite)
	self.image = self.img[colortype]
	
	# Follow the square around without need of an update() method:
        self.rect = square.rect


class MapFade(graphic.Graphic): #DON'T USE
    """Introduces the board.
    
    This is a huge sprite that covers the whole MAP_RECT in black.
    One at a time please.
	"""

    render_layer = 6
    render_clip = MAP_RECT
    update_auto = 1

    tt_text = None

    # TODO Perhaps replace with a surfarray 
    # routine, which would look hella better
    
    yspeed = -20


    def __init__(self, fadein=1):
	
	graphic.Graphic.__init__(self)
	self.fadein = fadein
	self.rect = Rect(MAP_RECT)
	self.image = pygame.Surface(self.rect.size)
	self.image.fill(MAP_BG_COLOR)
	self.image.set_alpha(120)	


    def update(self):
	
	self.rect.move_ip(0,self.yspeed)
	if not self.rect.colliderect(MAP_RECT): self.kill()
    

#######################################################################################
####### Miniature map graphics

class MiniSquare(graphic.Graphic):
    """Small representation of its corresponding Square instance.
    
    Used to control the locations of MiniSprite instances.

    Attributes:
    (class)
	offset -- pair of ints  -- position correction
	"""

    render_layer = 0
    render_clip = MINIMAP_RECT

    tt_text = None
    
    def __init__(self, square):
	
	graphic.Graphic.__init__(self)
	square.minisquare = self
	# Find place
	pos = (MINI_SQ_LENGTH * square.coord[0], MINI_SQ_LENGTH * square.coord[1])
	self.rect = MINI_PIECE_RECT.move(pos)
	self.rect.move_ip(self.offset)


class MiniLimboSquare(MiniSquare):
    """Small representation of its corresponding LimboSquare instance."""
    #different image


class MiniEntryExit(MiniSquare):
    """Small representation of its corresponding EntryTriangle or ExitArrow instance.
    """

    def __init__(self, tri):
	
	graphic.Graphic.__init__(self)
	# Facing
	self.image = pygame.transform.rotate(self.image, -90 * tri.entrypoint.efacing)
	# Find place
	pos = (MINI_SQ_LENGTH * tri.coord[0], MINI_SQ_LENGTH * tri.coord[1])
	self.rect = self.image.get_rect().move(pos)
	self.rect.move_ip(self.offset)

    
class MiniSprite(graphic.Graphic):
    """Small representation of its corresponding Piece instance.
    
    The arg passed to the constructor must have a .square attribute, anyway,
    and a .mini_image.
    
    Attributes:
    (instance)
	piece -- a Piece instance -- we represent
	"""
    #must have a MINI_SQ_LENGTH-sized self.image

    render_layer = 1
    render_clip = MINIMAP_RECT
    update_auto = 1 #?

    tt_text = None    
    
    def __init__(self, piece):
	
	self.piece = piece
	self.image = piece.mini_image
	graphic.Graphic.__init__(self)
	

    def update(self):
	
	if self.piece.alive(): self.rect = self.piece.square.minisquare.rect
	else: self.kill()
	

class LittleMapBox(graphic.Graphic):
    """A decoration that shows what part of the board is visible in MAP_RECT.

    1 exists always during normal play.
    For self.image, makes a rectangle the right size, dependent on MAP_RECT size.
    
    Attributes:
	board	    -- an objects.board.Board instance
	"""
	
    render_layer = 6
    render_clip = MINIMAP_RECT
    update_auto = 1 #!?

    tt_text = None
    
    def __init__(self, board):
	
	self.board = board
        graphic.Graphic.__init__(self)
	# Find the size (adding 1px border)
	size = (int(MAP_RECT.width / MINI_SCALE) + 2, int(MAP_RECT.height / MINI_SCALE) + 2)
	# Make image (filled with BLACK)
	self.image = pygame.Surface(size)
	self.rect = self.image.get_rect()
	# Draw box
	pygame.draw.rect(self.image, LITTLEMAPBOX_BOX_COLOR, self.rect, 1)
	# Make transparent
	self.image.set_colorkey(BLACK, RLEACCEL) #hmm
	
	
    def update(self):
	
	# Makes sure distance from self.rect.center to MINIMAP_RECT.center is in scale
	# with that for board.rect / MAP_RECT
	
	self.rect.centerx = MINIMAP_RECT.centerx + \
	int( ((MAP_RECT.centerx - self.board.rect.centerx)/MINI_SCALE) )
	self.rect.centery = MINIMAP_RECT.centery + \
	int( ((MAP_RECT.centery - self.board.rect.centery)/MINI_SCALE) )


##############################################################################
####### Stuff that crops up in HUD / MAP_RECT

class SectionLine(graphic.BoardGraphic):
    """The line that divides 2 board sections.
    
    Facing is UP or RIGHT.
    If facing is RIGHT, we're vertical, to the left of our square.
    If it is UP, we're horizontal, and to the bottom of our square.
    
    This class needs a <=3 * SQ_LENGTH wide, 2px or 4px high image to look right,
    with a transparent band SQ_LENGTH wide in the center.
	"""

    render_layer = 0
    render_clip = MAP_RECT
    update_auto = 0

    tt_text = None
    
    
    def __init__(self, square, facing):
	
	graphic.BoardGraphic.__init__(self)
	if facing != UP: self.image = pygame.transform.rotate(self.image, -90)
	self.rect = self.image.get_rect()
    	if facing != UP: 
	    self.rect.centery = square.rect.centery
	    self.rect.left = square.rect.left - int(self.rect.width / 2)
	else:
	    self.rect.centerx = square.rect.centerx
	    self.rect.bottom = square.rect.bottom + int(self.rect.height / 2)
	

class Marker(graphic.BoardGraphic):
    """Thing that attaches to a Piece instance.

    Parent class.

    Piece instances are responsible for update()ing and kill()ing these.
    
    Attributes:
    (instance)
	piece	-- the Piece
    (class)
	name	-- string
	_offset	-- (int, int) tuple -- how far down/right of piece center to display
	"""

    render_layer = 6 #above selectdec
    render_clip = MAP_RECT
    update_auto = 0

    tt_text = lambda self: self.name
 
    _offset = (0, 0) #default

    
    def __init__(self, piece):
	
	graphic.BoardGraphic.__init__(self)
	self.piece = piece
	self.rect = self.image.get_rect()
	self.rect.center = (self.piece.rect.centerx + self._offset[0], 
				self.piece.rect.centery + self._offset[1])

	#print >> DEBUG_STREAM, "Marker --> init, class:", self.__class__.__name__ #DEBUG
	
	
    def update(self):
	
	# This is called in _post_action() of Piece
	self.rect.center = (self.piece.rect.centerx + self._offset[0], 
			    self.piece.rect.centery + self._offset[1])


class OWMarker(Marker):
    """Shows piece is overwatching."""
    
    name = "Overwatching"
    _offset = (int(SQ_LENGTH * 0.4), int(SQ_LENGTH * 0.4))


class JamMarker(OWMarker):
    """Shows a gun is jammed."""

    name = "Jammed"
    

class ConvertMarker(Marker):
    """Shows that a blip is converting, and how many stealers are left to
    deploy.
    
    Attributes:
    (class)
	img -- int:surface dict -- 2 images for different values (1,2,3)
	"""

    tt_text = lambda self: "%s converting, %d Stealers remaining"%(self.piece.name, self.piece.value)

    name = "Convert marker" # not used
    _offset = (int(SQ_LENGTH / 3), int(SQ_LENGTH / 3))


    def __init__(self, piece):

	self.image = self.img[piece.value]
	Marker.__init__(self, piece)

    
    def update(self): #called especially
	
	if self.piece.value < 1: 
	    self.kill()
	else: 
	    self.image = self.img[self.piece.value]
	    Marker.update(self)


class DamageMarker(Marker):
    """For C_A_T"""

    name = "Damaged"
    _offset = (-int(SQ_LENGTH * (1/3)), -int(SQ_LENGTH * 0.4))
    

##############################################################################
####### Game Sprites

class Square(graphic.BoardGraphic):
    """Square of the board.

    Many may exist.  Acts as a holder for features/pieces using sprite groups.
    
    Attributes:
    (class) 
	offset
	corridor_image	-- default image
	room_image	-- image for room squares
	all_squares	-- pygame.sprite.Group -- all Square instances
    (instance)
	coord		-- pair of ints -- 'game-logical' coords
	isentry		-- bool -- stealer entry square
	efacing		-- int -- 0-3 -- facing off the board for an entry or exit
	isexit		-- bool -- marine exit square
	(e		-- entrytriangle instance -- if isexit or isentry)
	isdeploy	-- bool -- marine deployment square
	deploy_atribute	-- arbitrary object (probably string) -- used by missions
	objectivenum	-- int or None -- objective #
	
	_adjacents	-- list of Square instances -- up to 8 neighbours
	
	section		-- pygame.sprite.Group -- a room (this is also a boolean marker for Square (rather than LimboSquare) instances)
	features	-- pygame.sprite.Group -- Features that are on self
	piece		-- pygame.sprite.GroupSingle -- Piece instances (up to 1) that are on self
	hilite		-- pygame.sprite.GroupSingle -- HiLite instances (up to 1) that are on self
	"""

    render_layer = 0
    render_clip = MAP_RECT

    tt_text = None
    
    isentry = None	
    isexit = None
    isdeploy = None	
    objectivenum = None	
    
    def __init__(self, coord, section):
	
        graphic.BoardGraphic.__init__(self, self.all_squares, section)
	self.coord = coord

	# Find place onscreen
	pos = (SQ_LENGTH * coord[0], SQ_LENGTH * coord[1])
        self.rect = PIECE_RECT.move(pos)
	self.rect.move_ip(self.offset)
	
	MiniSquare(self)
	
        self.section = section
        self.piece = pygame.sprite.GroupSingle()
	self.hilite = pygame.sprite.GroupSingle()
        self.features = pygame.sprite.Group()
	self._adjacents = None #will become list


    def __repr__(self):

        return "<Square at %s>"%(self.coord,) #hmm
    
	
    def kill_hilite(self):
	"""kill_hilite() -> None

	Kill all hilites on self, if any."""

	for h in self.hilite.sprites(): h.kill()
    
	
    def get_adjacents(self):
        """get_adjacents() -> list
	
	Returns the list (necessarily nonempty) of Square instances which are adjacent."""

	# Must not be called until the whole board has been built
	
	if self._adjacents is None: self._calc_adjacents()
	return self._adjacents
	
    
    def _calc_adjacents(self):
	"""_calc_adjacents() -> None
	calculate adjacent squares"""
	
	# Should _adjacents be a spritegroup?
        borderrect = self.rect.inflate(2,2)
	self._adjacents = filter(lambda s: (s is not self and borderrect.colliderect(s.rect)), self.all_squares.sprites())


    def get_flaming(self):
	"""get_flaming() -> int

	Returns the # of Flames instances on self."""
	
	return len([x for x in self.features.sprites() if isinstance(x, Flames) ])


    def find_piece_near(self, team, maxdist):
        """find_piece_near(team, maxdist) -> (Piece instance, distance) or None
        get the nearest piece in a given team

        Distance is 0 if the piece is on self, 1 if on an adjacent square, etc.
        
        Returns nearest piece of those of given team within maxdist squares of square - or None if none.
        Used for Stealer and Blip to see if they may unlurk.
        Also used by AI0 (in progress)"""
        
        # First, see if it's even possible that there is an enemy piece near
        if self.rect.inflate((SQ_LENGTH * 2 * maxdist,) * 2).collidelist([piece.rect for piece in team.sprites()]) == -1: #i.e. no collisions
            return
        dist = 0	    #tree depth
        list = [self] #squares we're 'at' in this level of the tree
        exclu = []	    #squares we've already travelled
        while True:
            for sq in list: 
                for piece in sq.piece.sprites():
                    if piece.team is team: return piece, dist
            exclu = exclu + list
            newlist = []
            for sq in list:
                for a in sq.get_adjacents():
                    if a not in exclu and a not in newlist: newlist.append(a)
            dist += 1
            if dist > maxdist or not newlist: return
            list = newlist

            
    def get_team_is_near(self, team):
	"""get_team_is_near(team) -> bool
	
	Returns True if there is a piece from the given team within 6 squares, counting self as 0."""
        #Needed? - TW
	
	return bool(self.find_piece_near(team, 6))


    def get_passable(self, flamenum=False, catmove=False):
        """get_passable(flamenum=False, catmove=False) -> bool
	
	Bool for whether square can be travelled by a piece.
	If flamenum is true, will permit travel on flaming squares.
	If catmove is true, permits travel on squares occupied by a Piece."""
	
	# Is square occupied?
        if self.piece and not catmove: return False
	# Are there any blocking Features (e.g. doors)?
        elif self.features:
            for f in self.features.sprites():
		if f.blocksmove or (not flamenum and isinstance(f, Flames)): return False
        return True  


    def get_routable_to(self, s2, catmove=False):
	"""get_routable_to(s2, catmove=False) -> bool
	
	Returns true if there is a route from self to adjacent square s2, taking into
	account the presence or otherwise of Flames on self.
	If we are moving a C.A.T. then Pieces will be ignored."""

	# Degenerate case:
	if self is s2: return True
	
	flamenum = self.get_flaming()
	
	# Check that s2 is passable from here
	if not s2.get_passable(flamenum, catmove): return False
	
	# Find adjacents
	a1, a2 = self.get_adjacents(), s2.get_adjacents()
	
	# We must be adjacent to s2
	if s2 not in a1: return False
	
	# Trivially true in orthogonal/degenerate case
	if functions.ang(self, s2, UP) in ( F, B, L, R ): return True
	
	# Detect the 'route' squares, if they exist - i.e. the ones adjacent to both self and s2
	for sq in a1:
	    if sq in a2 and sq.get_passable(flamenum, catmove): return True #we only need 1
	return False


    def losthru(self):
        """losthru() -> bool
	
	Tells us whether square can be seen *through* (not to) by a piece."""

        if self.piece and not self.piece.sprites()[0].prone: return False #prone pieces do not block los
        for f in self.features.sprites():
            if f.blockslos: return False
        return True
    
    
    def flame(self, spread=SPREAD):
	"""flame(spread=SPREAD) -> None
	
	Sets up a Flames sprite on self, and has a go at killing any piece on self.  Returns None.
	If passed no args or SPREAD, will flame(SPREAD) orthogonally adjacent squares.
	spread is NOSPREAD on a selfdestruct by a Flamer, to save cycles(!)"""

	# No effect if there's a closed Door about
	for f in self.features.sprites():
	    if (isinstance(f, Door) and f.closed): return
	# Kill pieces etc
	for piece in self.piece.sprites():
	    if piece.fl_kill_scorereq:
		d = textual.DisplayDie(piece.team)
		if d.pips >= piece.fl_kill_scorereq: piece.kill(FLAMBEED)
	for feature in self.features.sprites():
	    if feature.fl_kill_scorereq:
		d = textual.DisplayDie()
		if d.pips >= feature.fl_kill_scorereq: feature.kill(FLAMBEED)
	# Add a flame feature
	Flames(self)
	# Spread
	if spread == SPREAD:
	    for square in self.get_adjacents():
		# We spread flame orthogonally within section
		if self.section.has(square) and functions.ang(self, square) in (F, L, B, R)\
		and square.get_flaming() < self.get_flaming():
		    square.flame(SPREAD)


class LimboSquare(Square): #hmm -- do I want to subclass off Square?
    """A pseudo-Square that pieces in entrytriangles use as their square.
    
    Attributes:
    (instance)
	entrytriangle -- EntryTriangle instance
    (class)
	all_limbosquares-- pygame.sprite.Group -- all LimboSquare instances
	"""

    def __init__(self, entrytriangle, number):
	
	graphic.BoardGraphic.__init__(self, entrytriangle.limbosquares, self.all_limbosquares)
	self.entrytriangle  = entrytriangle    
	self.section	    = None #marks out non-Square
	self.hilite	    = pygame.sprite.GroupSingle()
	self.features	    = pygame.sprite.Group() #not used!
	self.piece	    = pygame.sprite.GroupSingle()
	self._adjacents	    = [self.entrytriangle.entrypoint] #important for routefinding
	self.rect	    = Rect(0, 0, SQ_LENGTH, SQ_LENGTH)
	
	# Find my position
	coord = tuple(entrytriangle.coord)
	rdist = -1
	fdist = 1
	while number > 2:
	    rdist += 1
	    number -= 3
	fdist += number
	for i in range(self.entrytriangle.entrypoint.efacing):
	    rdist, fdist = fdist, -rdist
	coord = (coord[0] + rdist, coord[1] - fdist)

	# Find place onscreen
	pos = (SQ_LENGTH * coord[0], SQ_LENGTH * coord[1])
        self.rect = self.rect.move(pos)
	self.rect.move_ip(self.offset)

	self.coord = coord
	
	MiniLimboSquare(self)
	

class EntryTriangle(graphic.BoardGraphic):
    """Where Blips and Stealers lurk if they are not on the board.
    
    Attributes:
    (class)
	all		-- pygame.sprite.Group -- my fellow entrytriangles (from sulk.py)
	limbo_number	-- the number of LimboSquares to make
    (instance)
	coord
	entrypoint	-- Square instance -- where we lead to
	limbosquares	-- pygame.sprite.Group -- LimboSquare instances
	isexit		-- bool -- true if this is really an ExitArrow
	"""
	
    #RULES I don't think we need 9 limbosquares.  The theoretical max pieces at
    # an entrypoint is 6 (3 blips permitted plus 3 stealers) although the stealer
    # rule does seem a bit pointless sometimes.  (TODO)
	
    render_layer = 0
    render_clip = MAP_RECT

    tt_text = None
    
    limbo_number = 9
    
    isexit = 0 #hmm

    def __init__(self, square):
	
	graphic.BoardGraphic.__init__(self, self.all)
	self.entrypoint = square
	square.e = self #hmm

	# Point opposite to the entrypoint facing
	self.image = pygame.transform.rotate(self.__class__.image, -90 * self.entrypoint.efacing)

	# Get position #FIXME stupid
	self.rect = self.image.get_rect()
	self.rect.centerx = self.entrypoint.rect.centerx + FACEMAP[self.entrypoint.efacing][0]
	self.rect.centery = self.entrypoint.rect.centery + FACEMAP[self.entrypoint.efacing][1]
	self.coord =\
	(
	self.entrypoint.coord[0] + FACE_COORDMAP[self.entrypoint.efacing][0],
	self.entrypoint.coord[1] + FACE_COORDMAP[self.entrypoint.efacing][1]
	)
	
	# Make limbo squares
	self.limbosquares = pygame.sprite.Group()
	for i in range(self.limbo_number): LimboSquare(self, i)

	MiniEntryExit(self)


class ExitArrow(EntryTriangle):
    """Where Marines can leave the board.
	"""

    # Different graphics, functionally different,
    # but we'll just use the code upstairs.

    isexit = True

    #TODO -- more limbosquares (like 12)?
    #TODO -- just a better class generally?
    


#######################################################################################
####### Features

class Feature(graphic.Graphic):
    """A generic game feature parent class.

    Attributes:
    (class)
	all_features	-- spritegroup
	cc_kill_scorereq -- int or None -- can be destroyed in close combat on roll of this value or better
	sh_kill_scorereq -- int or None -- can be destroyed by shooting on roll of
	fl_kill_scorereq -- int or None -- can be destroyed when flamed on roll of
				(If these values make no sense (are 0, None or -ve) feature is immune.)
	type		-- string
	blocksmove	-- bool
	blockslos	-- bool
	persist		-- bool -- die in endphase?
    (instance)
	square		-- Square instance
	facing		-- int
	"""
	
    render_layer = 1
    render_clip = MAP_RECT

    tt_text = None #?

    persist = True	    #default
    blocksmove = True	    #default
    blockslos = True	    #default
    cc_kill_scorereq = 6    #default i.e. can be destroyed in close combat on roll of 6
    sh_kill_scorereq = 6    #default i.e. can be destroyed by shooting on roll of 6
    fl_kill_scorereq = None #default i.e. cannot be affected by flames
    sh_kill_scorereq = 6    #like a door
    aburst_kill_scorereq = 5
    aauto_kill_scorereq = 3
    mal_kill_scorereq = 4
    gr_kill_scorereq = 4
    
    
    def __init__(self, square, facing=UP):
	
        graphic.Graphic.__init__(self, self.all_features, square.features)
	
	# Feature facings never change
        self.facing = facing
        self.image = pygame.transform.rotate(self.image, -90 * self.facing)

	# Establish rect identity, so will follow square around.  
	# (all features must have SQ_LENGTHed (or smaller) images though)
        self.rect = square.rect
        self.square = square
	
	
    def kill(self, style=SHOT):
	"Rather like Piece.kill()"
	
	#TODO: use style/sounds
	graphic.Graphic.kill(self)

 
class Door(Feature):
    """Can be opened/closed by other pieces.

    Initial state is always closed.
    If opened, fires may spread.
    
    Attributes:
    (instance)
	closed -- bool
    (class)
	closesound -- sound
	opensound -- sound
	openimage -- Surface
	"""

    render_layer = 4 #same as Piece (not inherited from Feature)
    type = DOOR

    
    def __init__(self, square, facing=UP):
	
	Feature.__init__(self, square, facing)
	self.close(silent=True)


    def can_close(self):
	"""can_close() -> bool
	return true if closing is possible"""
	
	# RULES-->what about flames?  The rulebook is ambiguous.
	# I say we can close on flaming squares
	
	#use self.square.passable?
	if self.square.piece: return False
	for f in self.square.features.sprites():
	    if f is not self and f.blocksmove:
		return False
	else: return True


    def close(self, silent=False):
	"""-> None
	block movement, change image and attributes
        
        If not silent, play a sound."""
	
	if not silent: self.closesound.play()
        self.closed = self.blocksmove = self.blockslos = True
	self.cc_kill_scorereq = 6
        self.sh_kill_scorereq = 6
        self.gr_kill_scorereq = 5
        self.mal_kill_scorereq = 4
        self.aauto_kill_scorereq = 3
        self.aburst_kill_scorereq = 5
        self.image = pygame.transform.rotate(self.__class__.image, -90 * self.facing)
	# Move movable features
	for f in self.square.features.sprites():
	    if f is not self and not isinstance(f, Flames):
		self._displace(f)


    def open(self, silent=False):
	"""open(silent=False) -> None
	
	Allow movement and flames through the square, change image.
	If not silent, play sound."""
	
	if not silent: self.opensound.play()
        self.closed = self.blocksmove = self.blockslos = False
	self.cc_kill_scorereq = None
        self.sh_kill_scorereq = None
        self.gr_kill_scorereq = None
        self.mal_kill_scorereq = None
        self.aauto_kill_scorereq = None
        self.aburst_kill_scorereq = None
        self.image = pygame.transform.rotate(self.openimage, -90 * self.facing)
	self._flame_when_opened()


    def _flame_when_opened(self):
	"""_flame_when_opened() -> None
	
	Check for local Flames sprites, and flame() self.square as necessary.
	The square will never be flamed more than once this way"""
	#RULES --> VERIFY this behaviour
	
	for square in self.square.get_adjacents():
	    #spread happens orthogonally
	    if functions.ang(self.square, square) in (F, B, L, R):
		for f in square.features.sprites():
		    if f.__class__ is Flames:
			self.square.flame() #this will now SPREAD
			return

    
    def _displace(self, feature):
	
	choosables = []
	for sq in self.square.get_adjacents(): 
	    if sq.piece: continue
	    for f in sq.features:
		if f.blocksmove: continue
	    choosables.append(sq)
	    
	if not choosables: raise SystemExit, "Error code: b0rk b0rk b0rk"
	
	newsq = random.choice(choosables)
	feature.square = newsq
	feature.rect = newsq.rect


    def kill(self, style=SHOT):
	
	Feature.kill(self, style)
	if self.closed: self._flame_when_opened()


class DestroyedFeature(Feature):

    cc_kill_scorereq = None
    sh_kill_scorereq = None
    fl_kill_scorereq = None
    gr_kill_scorereq = None
    mal_kill_scorereq = None
    aauto_kill_scorereq = None
    aburst_kill_scorereq = None
    blockslos = blocksmove = False
    persist = True

    type = ""
    
    def __init__(self, afeature):
      
        self.image = afeature.destroyed_image
        Feature.__init__(self, afeature.square, afeature.facing)


class DestroyableFeature(Feature):
    
    def kill(self, style=SKEWERED):
	
	Feature.kill(self, style)
	DestroyedFeature(self)
        
        
class Ducting(DestroyableFeature):
    """Represents special ducting.  
    
    Stealers attack these in mission 6."""

    cc_kill_scorereq = 6
    sh_kill_scorereq = None
    fl_kill_scorereq = 1
    gr_kill_scorereq = 5
    mal_kill_scorereq = 4
    aauto_kill_scorereq = 4
    aburst_kill_scorereq = 5

    blocksmove = blockslos = True
    
    persist = True
    type = DUCTING


class Cryogenic_Tank(DestroyableFeature):
    """A Cryogenic Tank
    """
    
    type = CRYOGENIC_TANK

    cc_kill_scorereq = 6
    sh_kill_scorereq = None
    fl_kill_scorereq = 1
    gr_kill_scorereq = 5
    mal_kill_scorereq = 4
    aauto_kill_scorereq = 4
    aburst_kill_scorereq = 5

    def kill(self, style=SKEWERED):
	
	Feature.kill(self, style)
	d = DestroyedFeature(self)
        d.blockslos = d.blocksmove = True


class Bulkhead(DestroyableFeature):
    """A Bulkhead"""
    
    render_layer = 4
    cc_kill_scorereq = 6
    sh_kill_scorereq = None
    fl_kill_scorereq = None
    gr_kill_scorereq = None
    mal_kill_scorereq = 5
    aauto_kill_scorereq = None
    aburst_kill_scorereq = None

    type = BULKHEAD
    
    
class Crate(Feature):
    """Turns into rubble if attacked successfully"""
    
    fl_kill_scorereq = 6
    cc_kill_scorereq = 6
    sh_kill_scorereq = 6
    gr_kill_scorereq = 5
    mal_kill_scorereq = 4
    aauto_kill_scorereq = 4
    aburst_kill_scorereq = 5
    type = CRATE
    
    def kill(self, style=SKEWERED):
	
	Feature.kill(self, style)
	Rubble(self.square, self.facing)

    
class Rubble(Feature):
    "Indestructible crap lying around"
    
    type = RUBBLE
    cc_kill_scorereq = None
    sh_kill_scorereq = None
    fl_kill_scorereq = None
    gr_kill_scorereq = None
    mal_kill_scorereq = None
    aauto_kill_scorereq = None
    aburst_kill_scorereq = None
    blocksmove = False
    blockslos = False
    #TODO ricochet rule


class Flames(Feature):
    """Represents burning corridor."""

    render_layer = 2 #this puts it in a different render_container.  Watch out!
    
    cc_kill_scorereq = None
    sh_kill_scorereq = None
    fl_kill_scorereq = None
    gr_kill_scorereq = None
    mal_kill_scorereq = None
    aauto_kill_scorereq = None
    aburst_kill_scorereq = None

    blocksmove = False #for pieces in burning sections already, that is
    persist = False #die in endphase
    type = FLAMES
    #TODO animation
