# Sulk - sprites - pieces

#*****************************************************************************
#
#                                   Sulk
#
#       Copyright (c) 2002 Toby Woodwark <zirtix@softhome.net>
#
# This program is free software.  See the file COPYING for details.
#
#*****************************************************************************

"""
Sulk actor classes.
"""

##############################################################################
####### Import / set up

## Modules
import pygame

## My modules
import util #for Null
from constants import *  
import functions
import graphic
import misc
import anim
import textual
import buttons


##############################################################################
####### Base class

class Piece(graphic.Graphic):
    """A generic game piece parent class.
    
    Attributes:
    (instance)
        name            -- string   -- individual name, optional
	square		-- misc.Square instance -- where I am
	facing		-- int	-- 0-3
	prone		-- bool -- true if lying on the floor NOT USED YET
	apremaining	-- int	-- APs left on the piece
	
	_clocks		-- pygame.sprite.Group -- for textual.APClock, textual.AmmoClock
        buttons		-- pygame.sprite.Group -- for buttons.Button
	anims		-- pygame.sprite.GroupSingle -- to hold PieceAnim instances
	possessions	-- pygame.sprite.Group -- things in my possession e.g. C.A.T.s
	_sel		-- pygame.sprite.GroupSingle -- for anim.SelectDec
	_markers	-- pygame.sprite.GroupSingle -- for misc.Marker

        aiflags         -- dict of string:object -- used by AI implementations
                        (currently includes: 
                            'blocking' -- bool -- true if stealer is in the way of others [for AI0]
                            
                        )
    (class)
	team		-- pygame.sprite.Group -- buddies
	type		-- string   -- type name, see constants/strings.py
	weapon		-- int == FIXME
	
	movesound	-- sound
	ccsound	        -- sound
	turnsound	-- sound
	diesound	-- sound (TODO: variations)
	
	ap		-- int -- Default APs
	_usedoor_cost	-- int -- AP cost of door use
	_cc_cost	-- int -- AP cost of close combat cc
        _ccbonusmap     -- int:int dict -- Describes cc bonus when opponent is in given direction
        _ccdicemap      -- int:int dict -- Describes no of dice for given close combat direction
	_movemap	-- int:int dict -- Describes AP cost for different moves
	_turnmap	-- int:int dict -- Turn costs
	
        _button_class_list -- classes tuple -- button classes to be made available
	
	_keybmovedict	-- dict -- movement keys    
	_keybturndict	-- dict -- turning keys
	
        fl_kill_scorereq    -- int -- score required on a D6 to kill me with heavy flamer
	sh_kill_scorereq    -- int -- score required on a D6 to kill me with storm bolter
        mal_kill_scorereq   -- int -- score required on a D6 to kill me with assault cannon malfunction
        aburst_kill_scorereq -- int -- score required on a D6 to kill me with assault cannon burst
        aauto_kill_scorereq -- int -- score required on a D6 to kill me with assault cannon autofire
        gr_kill_scorereq    -- int -- score required on a D6 to kill me with plasma grenade
	"""

    render_layer = 4
    render_clip = MAP_RECT

    name = ""
    type = ""

    _keybmovedict = { K__F:F, K__F_R:F_R, K__R:R, K__B_R:B_R, K__B:B, K__B_L:B_L, K__L:L, K__F_L:F_L }
    _keybturndict = { K__LEFT:LEFT, K__RIGHT:RIGHT, K__DOWN:DOWN }

    _ccbonusmap = { F:0, L:0, R:0, B:0 }
    _ccdicemap = { F:1, L:1, R:1, B:1 }
    _movemap = { F_L:1, F:1, F_R:1, L:1, R:1, B_L:1, B:1, B_R:1 }
    _turnmap = { LEFT:0, RIGHT:0, DOWN:0 }
    
    fl_kill_scorereq = 2	    
    sh_kill_scorereq = 6
    aburst_kill_scorereq = 5
    aauto_kill_scorereq = 3
    mal_kill_scorereq = 4
    gr_kill_scorereq = 4
    
    _usedoor_cost = 1 
    _cc_cost = 1	    
    prone = False	    
    weapon = None
   
   
    def tt_text(self):
	
	if self.apremaining == 0:
	    apstring = "No APs"
	elif self.apremaining == 1:
	    apstring = "1 AP"
	else:
	    apstring = "%d APs"%self.apremaining
	    
	if self.name:
            name = "%s (%s)"%(self.name, self.type) 
        else:
            name = self.type
            
        text = "%s\n%s remaining\nStatus: %s"%\
	    (name, apstring, self.status)

	if self.possessions:
	    text += "\nPossessions:"
	    for p in self.possessions.sprites():
		text += "\n    " + p.tt_text()

	return text
    
    
    def __init__(self, square, facing):
        
        graphic.Graphic.__init__(self, self.team, square.piece,self.team.game.pieces)
	
        self.facing	= facing
        self.image	= pygame.transform.rotate(self.image, -90 * self.facing)
        self.rect	= square.rect
        self.square	= square
        self.status	= INACTIVE
        self.apremaining= self.ap

        self.aiflags    = {}

	# Create spritegroups for attachables
	self._markers	= pygame.sprite.GroupSingle()
	self._sel	= pygame.sprite.GroupSingle()
	self.anims	= pygame.sprite.GroupSingle()
	self._clocks	= pygame.sprite.Group()
	self.buttons	= pygame.sprite.Group()
	self.possessions= pygame.sprite.Group()
	
	# Show self on MINIMAP
	misc.MiniSprite(self)


#### tests
    def get_lurking(self):
	"""get_lurking() -> bool
	return true if piece is in limbo"""

	return (isinstance(self.square, misc.LimboSquare))
	

#### UI
    def build_buttons(self):
	"""build_buttons() -> None
	make pressable ui elements"""
	
	self.kill_buttons() #?
	for buttonclass in self._button_class_list: 
	    buttonclass(self)


    def kill_buttons(self):
	"""kill_buttons() -> None
	remove pressable ui elements"""
	
	for button in self.buttons.sprites(): 
	    button.kill()


    def build_visuals(self):
	"""build_visuals() -> None
	make the selection circle, etc"""
	
	self.kill_visuals() #overkill?
	self._sel.add(anim.SelectDec(self))
	self._clocks.add(textual.APClock(self))


    def kill_visuals(self):
	"""kill_visuals() -> None
	take down visuals"""

	if hasattr(self, '_sel'):
	    for selectdec in self._sel.sprites(): selectdec.kill()
	    for clock in self._clocks.sprites(): clock.kill()


    def keyb(self, key):
	"""keyb(key) -> None
	process a keypress"""
	
	if key in self._keybmovedict.keys():
	    for sq in self.square.get_adjacents():
		if functions.ang(self.square, sq, UP) == self._keybmovedict[key]:
		    goahead = self.can_move(sq)
		    if goahead is not None and self.get_points_avail() >= goahead:  
			self.take_points(goahead)
			self.move(sq)
		    return
	elif key in self._keybturndict.keys():
	    way = self._keybturndict[key]
	    goahead = self.can_turn(way)
	    if goahead is not None and self.get_points_avail() >= goahead:  
		self.take_points(goahead)
		self.turn(way)


#### APs
    def refresh(self):
	"""refresh() -> None
	make piece usable"""
	
	self.apremaining = self.ap
	self.status = INACTIVE


    def _post_action(self, actiontype):
	"""_post_action(actiontype) -> None

	What we do after every action (every thing costing APs).
	'actiontype' passed should be an int (defined in constants)
	and is passed directly to mission.post_action_script"""
	#N.B. Should be very careful about how this is called in subclasses.
	
	self.buttons.update()
	self.possessions.update()
	if self._markers: self._markers.sprite.update()
	#if self._sel: self._sel.sprite.update()
	# Keep track of last acting piece (for interrupts)
	self.team.game.set_actor(self)
	# Mission rule
	self.team.game.mission.post_action_script(self, actiontype)


    def get_points_avail(self):
	"""get_points_avail() -> int
	get the number of APs / CPs available for use by self"""

	#default (stealer) behaviour
	return self.apremaining


    def take_points(self, points):
	"""take_points(points) -> None
	pay some action points
	
	Works out which points the player wants to pay with and docks them from a piece / the cpclock."""
	
	#default (stealer) behaviour
	self.apremaining -= points

	
    def set_active(self, oldpiece=None):
	"""set_status(oldpiece=None) -> None
	set my status and the status of the previous active piece (if applicable)
	
	Called when I become 'cpiece' for an action phase instance, hopefully."""
	    
	if oldpiece is not None: 
	    # Do some 'so-called sanity' checks
	    if not self.team.has(oldpiece):
		print >> DEBUG_STREAM, "Piece --> set_active, oldpiece not in my team! SHOULD NOT HAPPEN" #DEBUG
	    if not oldpiece.alive():
		print >> DEBUG_STREAM, "Piece --> set_active, oldpiece is dead! SHOULD NOT HAPPEN" #DEBUG

	    #print >> DEBUG_STREAM, \
	    #"Piece --> set_active, oldpiece - status: %s, apremaining: %d"%\
	    #(STATUSNAMES[oldpiece.status], oldpiece.apremaining) #DEBUG
	    #print >> DEBUG_STREAM, \
	    #"Piece --> set_active, self - status: %s, apremaining: %d"%\
	    #(STATUSNAMES[self.status], self.apremaining) #DEBUG

	    if oldpiece.status == ACTIVE and oldpiece.ap == oldpiece.apremaining\
	    and (not isinstance(oldpiece, Genestealer) or oldpiece.just_turned is None):
		# The piece wasn't used.
		# N.B. Genestealers who have just acted for free may not return to INACTIVE.
		oldpiece.status = INACTIVE
		print >> DEBUG_STREAM, "Piece --> set_active, oldpiece set INACTIVE" #DEBUG
	    elif oldpiece.status == INACTIVE:
		# Nothing to do - piece wasn't even selected (shouldn't much happen)
		print >> DEBUG_STREAM, "Piece --> set_active, oldpiece remains INACTIVE" #DEBUG
		pass
	    else:
		# Normal case, the piece has acted this turn
		print >> DEBUG_STREAM, "Piece --> set_active, oldpiece set USED" #DEBUG
		oldpiece.status = USED
		
	#print >> DEBUG_STREAM, "Piece --> set_active, self set ACTIVE" #DEBUG   
	self.status = ACTIVE
	

#### Door
    def can_use_door(self, door):
        """can_use_door(door) -> int or None
	
	Returns AP cost or None."""
	
        sqr = door.square
        if (door.closed or door.can_close())\
	and sqr in self.square.get_adjacents() and self._arc_to(sqr) in FIREARC: 
            return self._usedoor_cost
	
	
    def use_door(self, door):
        """use_door(door) -> None
	open/close a door"""
	
        if door.closed: door.open()
        else: door.close()
	self._post_action(USING)
	

#### Turns
    def can_turn(self, way):
        """can_turn(way) -> int or None
	see how much a turn costs
	
	Returns the AP cost of the turn (can be zero!) or None if impossible."""
	
        return self._turnmap[way]         
    
    
    def turn(self, way, deployturn=False):
        """turn(way, deployturn=False) -> None
	rotate piece  
	
	If 'deployturn' bool is true, this is no action."""
	
	self.facing = (self.facing + way) % 4
        self.image = pygame.transform.rotate(self.__class__.image, -90 * self.facing)
	if not deployturn: 
	    self.turnsound.play()
	    self._post_action(TURNING)
    

#### LOS
    def can_see(self, square):
        """can_see(square) -> bool
	see if LOS is possible to a target square"""
	
	return (self._arc_to(square) in LOSARC and functions.seeable(self.square, square))
	
	
#### Movement
    def can_move(self, square):
        """can_move(square) -> int or None
	get the AP cost of a move"""
	
        if self.square.get_routable_to(square):
	    return self._movemap[functions.ang(self.square, square, self.facing)]        
	
	
    def move(self, square):
        """move(square) -> None
	change piece position"""
	
	self.movesound.play()
        self.square.piece.empty()
        self.square = square
        square.piece.add(self)
        self.rect = square.rect
	self._post_action(MOVING)
	

### Close combat
    def cc_score_is_max(self, target, score):
        """-> bool
        return True if the score is the max possible in this situation"""
        
        return (6 + self._ccbonusmap[self._arc_to(target.square)]) == score

    
    def fumble_possessions(self):
	"""fumble_possessions() -> None
	possibly fumble and drop any possessions"""
	
	map(lambda p: p.get_fumbled(), self.possessions.sprites())
	    

    def can_cc(self, target):
        """-> int or None
	get AP cost of close combat to target
	
	Return 0 or None if impossible."""
	
        if not self.team.has(target) and target.square in self.square.get_adjacents()\
	and self._arc_to(target.square) == F\
	and (self.team.enemy_team.has(target) or getattr(target, 'cc_kill_scorereq', False)): 
	    return self._cc_cost

        
    def reroll_one_cc_die(self, target):
        """-> int
        supply a new cc score"""
        
        return (textual.DisplayDie().pips + self._ccbonusmap[self._arc_to(target.square)])

    
    def get_cc_scores(self, target):
        """-> list
        supply a list of cc scores"""
        
        arc = self._arc_to(target.square)
        bonus = self._ccbonusmap[arc]
        numdice = self._ccdicemap[arc]
        scores = []
        for x in range(numdice):
            d = textual.DisplayDie(self.team)
            scores.append(d.pips + bonus)
	return scores
    
    
    def get_best_parry_result(self,sbest,target,tbest,tscores):
        """-> int"""
        
        return tbest

    
    def cc(self, target):
	"""cc(target) -> None
	fight a round of close combat against target 
	
	(Target must be a Piece or Feature instance.)"""
	
	# The target must be a piece or killable feature
	assert hasattr(target, '_ccdicemap') or hasattr(target, 'cc_kill_scorereq')
	
	self.ccsound.play()
  
	sscores = self.get_cc_scores(target)
        sbest = max(sscores)
	
	if hasattr(target, '_ccdicemap'):
	    # We're dealing with a Piece instance
	    
	    # Get rid of overwatch
	    if getattr(target, 'overwatching', False):
                target.overwatchoff()
            
	    # The target rolls dice
	    tscores = target.get_cc_scores(self)
            tbest = max(tscores)
	    tbest = self.get_best_parry_result(sbest, target, tbest, tscores)
            sbest = target.get_best_parry_result(tbest, self, sbest, sscores)
            
            # Who has the highest score?
            if sbest > tbest:
                target.kill(SKEWERED)
		self.fumble_possessions()
            elif sbest < tbest:
                if target.can_cc(self):
                    self.kill(SKEWERED)
		else:
		    target.turn((self.facing - target.facing + 2) % 4, deployturn=True)
		    self.fumble_possessions()
		target.fumble_possessions()
		    
	    else: #both still living:
		#hmm, code copying
		self.fumble_possessions()
		target.fumble_possessions()
		# Rotate enemy to face me - RULES optional for enemy FIXME
		target.turn((self.facing - target.facing + 2) % 4, deployturn=True)
		# RULES --> should stealers which turn to face their attacker be overwatched? hmm
		
	else: # We're attacking a feature:
	    self.fumble_possessions()
            if sbest >= target.cc_kill_scorereq: target.kill(SKEWERED)
	    
	self._post_action(CCING)
	    
	   
#### Misc

    def kill(self, style=SHOT):
	"""kill(style=SHOT) -> None
	remove piece from all sprite groups
	
	Play a sound appropriate to the way death occurred."""
	
	if self.alive():
	    self.kill_visuals()
	    self.kill_buttons()
	    
	    #TODO implement different sounds, in a better way
	    if style not in (SILENT, NODEATH): 
		snd = self.diesound.play(1, 3300)
		if snd: snd.fadeout(4600)
	    if style != NODEATH: 
		if isinstance(self, Blip):
		    self.team.casualties += self.value
		else:
		    self.team.casualties += 1

	    for p in self.possessions.sprites():
		p.dead_drop()
	    
	    # hmm, gc
	    # Private attributes can be safely disposed of - possibly
	    #if hasattr(self, "_markers"):
	    #    del self._markers
	    #if hasattr(self, "_clocks"):
	    #    del self._clocks
	    #if hasattr(self, "_sel"):
	    #    del self._sel
	else: 
	    print >> DEBUG_STREAM, "Piece --> kill --> dead already" #DEBUG
        for m in self._markers.sprites():
	    m.kill()
        self.aiflags = {}
	graphic.Graphic.kill(self)
       
	    
    def _arc_to(self, square):
	"""_arc_to(square) -> int or None
	return angle to target square
    
	Result is None if meaningless, otherwise it will be an int from either
	NOARC or LOSARC.  Some values in LOSARC are also in FIREARC.
	See functions.ang and constants for details."""
    
	return functions.ang(self.square, square, self.facing)


#### go...()

    def goCC(self, target):
	"""goCC(target) -> bool
	try to cc target, return true if an action is performed"""
	
	if self.get_points_avail() >= self._cc_cost and self.can_cc(target) is not None:
	    self.take_points(self._cc_cost)
	    self.cc(target)
	    return True
	else: return False
	
	
    def goHostile(self, target):
	"""goHostile(target) -> bool
	try to cc or shoot target, return true if an action is performed"""
	
	#default (Genestealer) behaviour = close assault
	return self.goCC(target)


    def goDoor(self, door, hostility=True):
        """goDoor(door, hostility=True) -> bool
	try to use a door, return true if an action is performed
	
	If 'hostility' is passed true, may cc this door."""
	
	if self.get_points_avail() >= self._usedoor_cost\
	and self.can_use_door(door) is not None:
	    self.take_points(self._usedoor_cost)
            self.use_door(door)
            return True
	elif hostility: return self.goHostile(door)
	else: return False
    
    
    def goMove(self, square):
	"""goMove(square) -> bool
	try to move to a square, return true on action"""
	
	goahead = self.can_move(square)
        if goahead is not None and self.get_points_avail() >= goahead:
            self.take_points(goahead)
            self.move(square)
	    return True
	return False
	
	
    def goSquare(self, square):
	"""goSquare(square) -> bool
	find an action for a square return true on action"""
	
        return self.goMove(square)


    def goTurn(self, way):
	"""goTurn(way) -> bool
	try to turn, return true on action"""
	
	goahead = self.can_turn(way)
        if goahead is not None and self.get_points_avail() >= goahead:
            self.take_points(goahead)
            self.turn(way)
	    return True
        else: return False


class ReloadsMixIn:
    """A piece that can reload its weapon."""

    _reload_cost = 4
    reloads = 1


    def __init__(self):

        self.reloadsremaining = self.reloads

        
    def goReload(self):
        """-> bool"""

        cost = self.can_reload()
        if cost is not None and cost <= self.get_points_avail():
            self.take_points(cost)
            self.reload()
            return True
        return False
        
        
    def can_reload(self):
        """-> int or None"""
        
        if self.reloadsremaining > 0:
            return self._reload_cost
    
    
    def reload(self):
        """-> None"""
        
        self.reloadsremaining -= 1
        self.ammoremaining = self.ammo
        self._post_action(RELOADING)

        
class ParryMixIn:
    """A piece that can parry in close combat"""
    
    _parry_arc = ( F ) #Default, can parry only forwards
    
    
    def get_best_parry_result(self, sbest, target, tbest, tscore):
        """get_best_parry_result(self, sbest, target, tbest, tscore) -> int
        decide whether or not to parry, and then do so
        
        Returns the new enemy score."""

        canparry = (self._arc_to(target.square) in self._parry_arc)

        if (sbest < tbest or target.cc_score_is_max(self, tbest)) and canparry:
            # The target has a better or maximum score, we might as well parry
            return self._parry(tscore, target)
        elif sbest > tbest:
            # The target has already lost, no point parrying
            return tbest
        elif canparry:
            self.answer = 0 
            self.team.player.UIinput.Parry(self.team.game.phase, self, tscore)
            while not(self.team.game._done or self.answer):
                self.team.game._game_loop()
            if self.answer:
                return self._parry(tscore, target)
            else:
                return tbest
        return tbest


    def _parry(self, scores, target):
        """_parry(scores, target) -> int
        carry out a parry against an opponent
        
        Returns the new maximum score."""
        
        #TODO - a 'parry' sound
        scores.remove(max(scores))
        scores.append(target.reroll_one_cc_die(self))
        return max(scores)
    


##############################################################################
####### Bad guys

class Genestealer(Piece):
    """A bad guy that charges about the Hulk at 6 APs each, can turn for free
    and has gnashing teeth.  Can survive a hard vacuum, but not hard bullets.
    
    Attributes:
    (instance)
	sees_me		    -- spritegroup
	just_turned	    -- int -- a facing
	may_unlurk_at_will  -- bool -- true if self has lurked for a turn already
    (class)
	_unlurk_cost	    -- an int -- cost to move out of a 'triangle'
	"""
        
    type = GENESTEALER
    
    ap = 6
    #3 dice, no bonus. 2 dice if cced from side or rear (RULES-->DEATHWING RULE)
    _ccbonusmap = { F:0, L:0, R:0, B:0 }
    _ccdicemap = { F:3, L:2, R:2, B:2 }
    _movemap = { F_L:1, F:1, F_R:1, L:2, R:2, B_L:2, B:2, B_R:2 }
    _turnmap = { LEFT:0, RIGHT:0, DOWN:1 }
    _button_class_list = ( buttons.AttackButton, buttons.UsedoorButton ) 
    
    mal_kill_scorereq = 4
    gr_kill_scorereq = 4
    _unlurk_cost = 1

    
    def tt_text(self):

	text = Piece.tt_text(self)
	if self.get_lurking():
	    if self.may_unlurk_at_will:
		text += "\nMay unlurk at will"
	    else:
		text += "\nMay not unlurk this turn"
	return text
	

    def __init__(self, square, facing):
	
        Piece.__init__(self, square, facing)
	self.just_turned = None
	self.sees_me = pygame.sprite.Group()

    
    def _post_action(self, actiontype):
	
	self.just_turned = None
	sees_me = self.sees_me
	# Before we update who can see me, check for blip LOS
	functions.discover_blips(sees_me.sprites())
	# Empty the grp
	sees_me.empty()
	# Now update who can see me
        self.update_sees_me()
        Piece._post_action(self, actiontype)


    def update_sees_me(self):
        """update_sees_me() -> None
        check for and cause overwatch fire"""
        
	for enemy in self.team.enemy_team.sprites():
	    if enemy.can_see(self.square):
		self.sees_me.add(enemy)
		if enemy.overwatching: 
		    enemy.goShoot(self, overwatchfire=True)
    
    
    def refresh(self):
	
	Piece.refresh(self)
	self.just_turned = None
	if self.get_lurking(): self.may_unlurk_at_will = True


    def keyb(self, key):
	
	# (Extra lurking rule)
	if self.get_lurking() and key in self._keybmovedict.keys(): 
	    self.goSquare(self.square.entrytriangle.entrypoint)
	else: Piece.keyb(self, key)


    def can_turn(self, way): 
	
	# Genestealers may not turn 90 degrees twice in a row for free, or turn if lurking:
	if not self.get_lurking():
	    if way == self.just_turned and way in (LEFT, RIGHT): 
		return self._turnmap[DOWN]
	    else: return self._turnmap[way] 


    def turn(self, way, deployturn=False):
	
	Piece.turn(self, way, deployturn)
	if not deployturn:
	    self.just_turned = way


    def can_cc(self, target):
	
	if self.get_lurking():
	    if self.may_unlurk_at_will \
	    and target.square is self.square.entrytriangle.entrypoint\
	    and (
		self.team.enemy_team.has(target)\
		or (isinstance(target, misc.Feature) and target.cc_kill_scorereq)
		): 
		return self._cc_cost
	else: return Piece.can_cc(self, target)


    def can_move(self, square):
	
	if self.get_lurking():
	    if square is self.square.entrytriangle.entrypoint and square.get_passable()\
            and (self.may_unlurk_at_will):# or not square.get_team_is_near(self.team.enemy_team)): #??? - TW
		 return self._unlurk_cost
	else: return Piece.can_move(self, square)

   	
class Blip(Genestealer):
    """A bad guy with a value in terms of stealers.
    
    Extra attributes:
	value	    -- int -- number of stealers inside
	revealedimg -- dict of value:Surface -- different images for non-secret blips
	"""
    
    type = BLIP

    _ccbonusmap = { F:0, L:0, R:0, B:0 } #but close combat SHOULD NOT HAPPEN
    _ccdicemap = { F:0, L:0, R:0, B:0 }
    _cc_cost = None			    #so can_cc always returns None.
    _movemap = { F_L:1, F:1, F_R:1, L:1, R:1, B_L:1, B:1, B_R:1 }
    _turnmap = {}                           #this says I *cannot* turn
    _button_class_list = ( buttons.UsedoorButton, buttons.ConvertButton )
    _keybturndict = {}	
    
    _post_action = Piece._post_action	    #not inherited from Genestealer
    can_turn = can_cc = util.Null	    #can never turn or cc
    
    
    def __init__(self, square, value):
	
	Genestealer.__init__(self, square, UP)
        self.value = value
	if self.team.secrets_showing: self.reveal()
	if self.get_lurking(): 
	    self.may_unlurk_at_will = \
	    (not square.entrytriangle.entrypoint.get_team_is_near(self.team.enemy_team)) #bool


    def reveal(self):
	"""reveal() -> None
	
	Display self.value."""
	
	self.image = self.revealedimg[self.value]
	
	
    def unreveal(self):
	"""unreveal() -> None
	
	Hide self.value."""
	
	self.image = self.__class__.image
	
	
    def can_use_door(self, door):
	
	# Do not inherit from Genestealer
	# No facing, and can't bring self into marine LOS
        if door.square in self.square.get_adjacents(): 
	    if door.closed:
		#pretend it's open...
		door.blockslos = False	    
		for marine in self.team.enemy_team.sprites():
		    if marine.can_see(self.square):
			door.blockslos = True	
			return		
		door.blockslos = True
	    elif not door.can_close(): return
	    return self._usedoor_cost
	
	
    def can_convert(self):
	"""can_convert() -> bool
	see if blip can convert
	
	Blips may only convert volutarily if they have taken no other actions."""
	
	return (self.apremaining == self.ap)
	
	 
    def convert(self, team=None):
	"""convert(team=None) -> None
	start placing Genestealers around this blip
	
	Team is the deploying team (assumed self.team)."""
	
	if team is None: team = self.team
	self.convertvoluntary = (team is self.team)
	self._markers.add(misc.ConvertMarker(self))
	
	# Magically start an inputhandler
	team.player.UIinput.Convert_A_Blip(self.team.game.phase, self)


    def can_move(self, square):
	
	# Blips may not move into marine LOS, or next to a marine:
	p = Genestealer.can_move(self, square)
	if p:
	    for marine in self.team.enemy_team.sprites():
		if marine.can_see(square) or square in marine.square.get_adjacents(): return
	    else: return p


    def kill(self, style=NODEATH): 
	
	# Different default style
	Piece.kill(self, style)
        
	
class Ambush_Counter(Blip):
    
    type = AMBUSH_COUNTER
    
    _button_class_list = (buttons.UsedoorButton,)
    
    reveal = unreveal = util.Null
    can_move = Piece.can_move #not inherited from Blip

    
    def try_to_jam_guns(self):
        
        for enemy in self.team.enemy_team.sprites():
            if enemy.can_see(self.square):
                if enemy.overwatching:
                    enemy.fire_at_empty_space()

                    
    def convert(self, team=None):
	"""convert(team=None) -> None
	start placing Genestealers around this blip
	
	Team is the deploying team (assumed self.team)."""
        
        if self.value:
            if team is None: team = self.team
            self.convertvoluntary = (team is self.team)
            self._markers.add(misc.ConvertMarker(self))
            # Magically start an inputhandler
            team.player.UIinput.Convert_A_Blip(self.team.game.phase, self)
        else:
            self.kill(NODEATH)
            self.try_to_jam_guns()

            
    def _post_action(self, actiontype):
	
	for enemy in self.team.enemy_team.sprites():
	    if enemy.can_see(self.square):
                self.convert(enemy.team)
                # This code is stale (TW) ...?
		#if enemy.overwatching: 
		#    enemy.goShoot(self, overwatchfire=1)
	Piece._post_action(self, actiontype)
        
        
    def can_use_door(self, door):
        """-> int or None
        
	Returns AP cost or None."""
        
        if (door.closed or door.can_close())\
	and  door.square in self.square.get_adjacents(): 
            return self._usedoor_cost

        
##############################################################################
####### Good guys

class Terminator_Marine(Piece):
    """Good guy parent class"""
    
    #TODO move more stuff in here
    type = TERMINATOR_MARINE
    
    _ccbonusmap = { F:0, L:0, R:0, B:0 }
    _ccdicemap = { F:1, L:1, R:1, B:1 }
    _movemap = { F_L:1, F:1, F_R:1, L:None, R:None, B_L:2, B:2, B_R:2 }
    _turnmap = { LEFT:1, RIGHT:1, DOWN:2 }
    _button_class_list = (  buttons.AttackButton, buttons.ShootButton,\
                            buttons.UsedoorButton) #3 actions
    
    ap = 4
    _lurk_cost = 1
    _pass_cat_cost = 1
    _pickup_cat_cost = 1
    _take_cat_cost = 1
    
    mal_kill_scorereq = 5
    gr_kill_scorereq = 6
    

class Terminator_Marine_with_Storm_Bolter(Terminator_Marine):
    """Good guy with a big gun.  Has 4 APs, the ability to fire on overwatch,
    and huge chunks of armor on him.  n33t.
    
    Attributes:
	squadname	-- string -- supplied by sulk.py at mission start
	weapon		-- int
	ammo		-- int or -1 (infinite)
	ammoremaining	-- int or -1 (infinite)
	_sh_cost	-- int -- cost of shoot action
	_ow_cost	-- int -- cost of going on overwatch
	_ow_weapon_range-- int -- gun range in overwatch
	_unjam_cost	-- int -- cost of unjamming
	_lurk_cost	-- int -- cost of using an ExitArrow
	
	_jammed_button_class_list -- tuple -- to replace _button_class_list

	_pass_cat_cost	-- int -- cost of passing a C.A.T. to adjacent marine
	_take_cat_cost	-- int -- cost of taking a C.A.T. from adjacent marine
	_pickup_cat_cost-- int -- cost of getting a C.A.T. from the floor

	_fire_bonus	-- int -- added to die rolls
	_max_fire_bonus	-- int -- max permitted value for _fire_bonus
	may_mns_shoot	-- bool -- self may shoot for free now
	try_mns		-- bool -- shoot for free when possible
	overwatching	-- bool -- shoot in stealer turn
	gunjammed	-- bool -- gun is inoperative
	_last_target	-- pygame.sprite.GroupSingle -- the last Piece in the crosshairs
	"""
        
    type = TERMINATOR_MARINE_WITH_STORM_BOLTER
    
    weapon = GUN #Storm Bolter
    ammo = ammoremaining = -1 #infinite
    _sh_cost = 1
    _ow_cost = 2
    _ow_weapon_range = 12
    _unjam_cost = 1
    
    _max_fire_bonus = 4 #so (2,2) is always a miss
    _button_class_list = (  buttons.AttackButton, buttons.ShootButton,\
			    buttons.MovenshootButton, buttons.OverwatchButton, buttons.UsedoorButton) #5 actions
    _jammed_button_class_list = (  buttons.AttackButton, buttons.UnjamButton, buttons.UsedoorButton) #3 actions (hmm)
    
    
    def __init__(self, square, facing, squadname=""):
    
	Piece.__init__(self, square, facing)
	self.squadname = squadname
	self._last_target = pygame.sprite.GroupSingle()
	self._fire_bonus = 0
	self.gunjammed = False
	self.overwatching = False
	self.try_mns = False
	self.may_mns_shoot = False
	self.ammoremaining = self.ammo


    def fire_at_empty_space(self):
        """-> None
        Called when fire in overwatch at an ambush_counter having value of 0 stealers
        """
        
        anim.Flash(self, GUN)
        d1, d2 = textual.DisplayDie(self.team).pips, textual.DisplayDie(self.team).pips
        if d1 == d2:
            self.overwatchoff()
            self.gunjam()

            
    def build_buttons(self):
	
	Piece.build_buttons(self)
	# Add a cat button if relevant
	if CAT in self.team.game.mission.FLAGS: 
	    buttons.CATButton(self)


    def keyb(self, key):
	
	if key in self._keybmovedict.keys() and self.square.isexit\
	and FACINGANGLES[self.square.efacing] == self._keybmovedict[key]:
	    # Move into lurkspot
	    for sq in self.square.e.limbosquares.sprites():
		if not sq.piece: lurksq = sq
		break
	    else: 
		# Full!
		return
	    self.goMove(lurksq)
	else: Piece.keyb(self, key)


    def refresh(self):
	
	if self.get_lurking(): return #?
	
	Piece.refresh(self)
	self._last_target.empty()
	self.may_mns_shoot = self.try_mns = self._fire_bonus = False
	if self.overwatching: self.overwatchoff()


    def _post_action(self, actiontype, discover=1):
	
	self.may_mns_shoot = False  #overridden in move() though
        self._fire_bonus = 0	    #overridden in shoot() though
	if discover: functions.discover_blips([self])
	Piece._post_action(self, actiontype)


#### APs
    
    def get_points_avail(self, takecat=False):
	
	# If takecat is true, permit the use of points by non-ACTIVE pieces
	
	if self.team.status == ACTIVE:
	    if self.team.game.cpbutton.status is PRESSED:
		if self.status != ACTIVE and not takecat: 
		    return self.team.game.cpclock.cpremaining
		else: 
		    return self.apremaining + self.team.game.cpclock.cpremaining #we can use both cps and aps
	    else: return self.apremaining
	else: #i.e. it is stealer turn:
	    return self.team.game.cpclock.cpremaining


    def take_points(self, points, takecat=False):

	# As above
	
	if points == 0: return
	elif self.team.status != ACTIVE:
	    #Genestealer turn - CPs only
	    self.team.game.cpclock.cpremaining -= points
	elif self.team.game.cpbutton.status == PRESSED:
	    #marine turn with the CPs on
	    if self.status is ACTIVE or takecat: 
		#we'll first exhaust the APs
		for i in range(points):
		    if self.apremaining: self.apremaining -= 1
		    else: self.team.game.cpclock.cpremaining -= 1
	    else: self.team.game.cpclock.cpremaining -= points
	else: Piece.take_points(self, points)


    def set_active(self, oldpiece=None):
	
	if self.team.game.cpbutton.status == PRESSED: #cruft
	    # Don't actually change any status
	    textual.ScrollMessage('Pieces keep status.')
	else:
	    Piece.set_active(self, oldpiece)


#### Shooting

    def can_shoot(self, target):
        """can_shoot(target) -> int or None
	
	If LOS, fire arc, overwatch status, and range permit shooting, returns cost of firing. 
	Returns None if firing is not permitted."""
	
	# Find cost of firing
	if self.overwatching: go = 0
	elif self.try_mns:
	    # We are trying to move and shoot, so we must be *able* to.
	    # (Cost is zero, when possible)
	    if self.may_mns_shoot: go = 0
	    else: 
		textual.ScrollMessage("Can't move and shoot.") #hmm rather not have a msg here
		return
	else: go = self._sh_cost #normal case
		
	# Can we fire?
	if target.sh_kill_scorereq and self.ammoremaining and not self.gunjammed\
	and self._arc_to(target.square) in FIREARC and functions.seeable(self.square, target.square)\
	and (
	    not self.overwatching or \
	    (self.team.status != ACTIVE and functions.find_range(self, target) <= self._ow_weapon_range)
	    ): 
		return go


    def shoot(self, target):
        """shoot(target) -> None
	
	Fires a shot at target (can be a feature or a piece)."""
	
	def killtarget():
	    "What we do when a kill occurs."
	    # See who has los to the expiring thing, (not incl self)
	    seelist = [m for m in self.team.sprites() if m.can_see(target.square)] #expensive...
	    # Kill the target
	    target.kill(SHOT)
	    # Check the los to blips, but for everybody (stronger than postaction)
	    functions.discover_blips(seelist)
	    print >> DEBUG_STREAM, "Terminator_Marine_with_Storm_Bolter --> shoot --> killtarget" #DEBUG  
	    
	# Go and shoot

	# Eyecandy, sounds	
        anim.Flash(self, GUN)

	# Roll dice
	fb = int(self._fire_bonus)
        d1, d2 = textual.DisplayDie(self.team).pips, textual.DisplayDie(self.team).pips
	print >> DEBUG_STREAM, "Terminator_Marine_with_Storm_Bolter --> shoot, firebonus:%d, dice:%d,%d"%(fb,d1,d2) #DEBUG
	
	# Find the score required to kill
	sr = int(target.sh_kill_scorereq)
	
        if self.overwatching and d1 == d2:
            self.overwatchoff()
	    self.gunjam()
	
	# Try to kill
	if not (self.try_mns or self.overwatching):
	    if self._last_target.has(target): sr -= fb 
	    if d1 >= sr or d2 >= sr:
		# Hit:
		killtarget()
                fb = 0
	    else: 
		# Miss:
		fb += 1
		if fb > self._max_fire_bonus: fb = self._max_fire_bonus
		self._last_target.add(target)
        else:
	    if d1 >= sr or d2 >= sr: 
		# Hit:
                killtarget()
                fb = 0
	    else: 
		# Miss:
                self._last_target.add(target)
		fb = 0

	# We must only become game.last_actor if we are *acting* (not overwatching)
	if not self.overwatching: self._post_action(SHOOTING, discover=False) # Blip discovery is done by killtarget()

	# Post-_post_action() jiggery pokery
	self._fire_bonus = fb


#### Misc

    def can_use_door(self, door):
	
	if not self.overwatching: return Piece.can_use_door(self, door)


    def use_door(self, door):
	
	#Extra blip stuff
	
	Piece.use_door(self, door) # This checks for my los
	
	if door.open: 
	    # Check whole team los to blips
	    functions.discover_blips([m for m in self.team.sprites() if m is not self and m.can_see(door.square)])

	
    def can_cc(self, target):
	
	if not self.overwatching: return Piece.can_cc(self, target)
	
    
    def can_turn(self, way):
	
	if not self.overwatching: return Piece.can_turn(self, way)
	
	
    def can_move(self, destsquare):

	if not self.overwatching: 
	    if isinstance(destsquare, misc.LimboSquare):
		# The limbo must be attached to an exit arrow
		if destsquare.entrytriangle.isexit and self.square is destsquare.entrytriangle.entrypoint:
		    #RULES how many APs should it cost?  The rulebook says nothing about facing,
		    #so I am going to say 1 regardless of facing
		    return self._lurk_cost
	    else:
		return Piece.can_move(self, destsquare)
	

    def move(self, destsquare):
	
	Piece.move(self, destsquare)
	# Post _post_action() jiggery pokery
	self.may_mns_shoot = True
	if self.get_lurking(): 
	    print >> DEBUG_STREAM, "Terminator_Marine_with_Storm_Bolter --> move, moved offboard" #DEBUG
	    self.status = USED


#### Overwatch

    def gunjam(self):
        """gunjam() -> None
	
	Jam the gun."""
	
	#N.B. never happens when I am active/selected
	self.jamsound.play()
        self.gunjammed = True
	self._markers.add(misc.JamMarker(self))
	self._button_class_list = self.__class__._jammed_button_class_list


    def ungunjam(self):
        """ungunjam() -> None
	
	Unjam the gun."""
	
	#self.unjamsound.play() FIXME
	self._button_class_list = self.__class__._button_class_list
	self.build_buttons()
        self.gunjammed = False
	self._markers.sprites()[0].kill()
	self._post_action(ACTING)


    def overwatchon(self):
        """overwatchon() -> None
	
	Puts on overwatch."""

	# RULES --> Technically marines may act when in overwatch, automatically cancelling it.
	# My implementation is to force the cancellation of overwatch before any action is permitted.
	
        self.overwatching = True
	self.try_mns = False
	self._markers.add(misc.OWMarker(self))
	self._post_action(OVERWATCHING)


    def overwatchoff(self):
        """overwatchoff() -> None
	
	Takes off overwatch."""
	
        self.overwatching = False
	self._markers.sprites()[0].kill()
	self.buttons.update()


#### go...()

    def goShoot(self, target, overwatchfire=False):
	"""goShoot(target) -> bool
	
	If overwatchfire bool is set, allow shooting just if we are in overwatch.
	Try to shoot target.
	Return True if it is done."""
    
	if (overwatchfire and self.overwatching)\
	or (
	    (not overwatchfire and not self.overwatching)
	    and (
		(self.try_mns and self.may_mns_shoot)\
		or self.get_points_avail() >= self._sh_cost #(normal case)
		)
	    ):
	    goahead = self.can_shoot(target)
	    if goahead is not None:
		self.take_points(goahead)
		self.shoot(target)
		return True
	return False


    def goHostile(self, target):

	#N.B. never used for overwatch.

	# First try shooting behaviour
	if self.goShoot(target): return True
	# Fall back to CC
	else: return Piece.goHostile(self, target)


    def goUngunjam(self):
	"""goUngunjam() -> bool
	
	Try to unjam.
	Return True if it is done."""
    
	if self.gunjammed and self.get_points_avail() >= self._unjam_cost:
	    self.take_points(self._unjam_cost)
	    self.ungunjam()
	    return True
	else: return False
	
#### C.A.T.
	
    def can_pickup_cat(self, cat):
	"""can_pickup_cat(cat) -> int or None"""

	if cat.square is self.square and not self.possessions.has(cat): 
	    return self._pickup_cat_cost


    def can_pass_cat(self, cat, piece):
	"""can_pass_cat(cat, piece) -> int or None

	Returns cost of passing cat to marine"""

	#RULES -- I am assuming you can only pass orthogonally
	
	if cat.alive() and self.possessions.has(cat) and self.team.has(piece)\
	and piece.square in self.square.get_adjacents()\
	and self._arc_to(piece.square) in (F, B, L, R):
	    return self._pass_cat_cost


    def can_take_cat(self, cat, piece):
	"""can_take_cat(cat, piece) -> int or None

	Returns cost of taking cat from marine"""

	if cat.alive(): return self._take_cat_cost


    def pickup_cat(self, cat):
	"""pickup_cat(cat) -> None"""
	
	if not cat.get_fumbled() and cat.alive(): 
	    cat.set_piece(self)
	self._post_action(ACTING)


    def pass_cat(self, cat, piece):
	"""pass_cat(cat, piece) -> bool
	
	Returns true if we made the piece try to take the cat"""

	#RULES If I succesfully pass the C.A.T. but taker fumbles
	# it, should it end up in his square or in mine?
	
	if not cat.get_fumbled() and cat.alive(): 
	    # The cat is out of my hands now (leaving it in my square)
	    #cat.set_piece()
	    cat.drop()
	    piece.take_cat(cat, self)
	    self._post_action(ACTING)
	    return True
	else:
	    self._post_action(ACTING)


    def take_cat(self, cat, piece):
	"""take_cat(cat, piece) -> None
	
	Take cat from piece.  Not an action"""
	
	if not cat.get_fumbled() and cat.alive():
	    # Yes, get the cat
	    cat.set_piece(self)
	cat.update()

    
    def goPickup_cat(self, cat):
	"""goPickup_cat(self, cat) -> bool"""
	
	cost = self.can_pickup_cat(cat)
	if cost is not None and self.get_points_avail() >= cost:
	    self.take_points(cost)
            self.pickup_cat(cat)
            return True
	else: return False


    def goPass_cat(self, cat, piece):
	"""goPickup_cat(self, cat) -> bool
	
	Self and the taking piece must *both* have APs available."""
	
	cost = self.can_pass_cat(cat, piece)
	tcost = piece.can_take_cat(cat, self)
	if cost is not None and self.get_points_avail() >= cost\
	and tcost is not None and piece.get_points_avail(takecat=True) - cost >= tcost:
	    self.take_points(cost)
	    if self.pass_cat(cat, piece):
		piece.take_points(tcost, takecat=True)
            return True
	else: return False


class Terminator_Marine_Sergeant(Terminator_Marine_with_Storm_Bolter):
    """Good guy with stripes.  Bish bash bosh.
    
    Gives +30 seconds to turnclock.
    
    Attributes:
    (class)
	timer_bonus
	"""
    
    type = TERMINATOR_MARINE_SERGEANT
    _ccbonusmap = { F:1, L:1, R:1, B:1 }
    timer_bonus = 30


class Terminator_Marine_Sergeant_with_Power_Sword(Terminator_Marine_Sergeant, ParryMixIn):
    """Req different graphics"""
    
    type = TERMINATOR_MARINE_SERGEANT_WITH_SWORD
    
    
class Terminator_Marine_with_Lightning_Claws(Terminator_Marine_with_Storm_Bolter):
    
    type = TERMINATOR_MARINE_WITH_LIGHTNING_CLAWS

    _button_class_list = (  buttons.AttackButton, buttons.UsedoorButton )
    goHostile = Piece.goHostile
    _ccbonusmap = { F:2, L:2, R:2, B:2 }
    _ccdicemap = { F:2, L:1, R:1, B:1 }


class Terminator_Marine_with_Thunder_Hammer(Terminator_Marine_with_Lightning_Claws, ParryMixIn):
    """Good guy with a fat-ass hammer"""
    
    type = TERMINATOR_MARINE_WITH_THUNDER_HAMMER
    
    
    _button_class_list = (  buttons.AttackButton, buttons.UsedoorButton, buttons.SelfdestructButton )
    _ccbonusmap = { F:2, L:0, R:0, B:0 }
    _ccdicemap = { F:1, L:1, R:1, B:1 }
    _parry_arc = ( F, L )
    _selfdestruct_cost = 1

    
    def can_selfdestruct(self):
	"""can_selfdestruct() -> int or None
	
	Return _selfdestruct_cost; else return None."""
	
	return self._selfdestruct_cost
	
	
    def selfdestruct(self):
	"""selfdestruct() -> None
	
	Kills every piece and feature in own section."""
	
	self.destructsound.play()
        for square in self.square.section.sprites():
            for s in square.piece.sprites(): s.kill(SILENT) #silent kills
            for f in square.features.sprites(): f.kill(SILENT) #even open doors are destroyed


    def goSelfdestruct(self):
	"""goSelfdestruct() -> bool

	Try to selfdestruct."""
	
	if self.get_points_avail() >= self._selfdestruct_cost and self.can_selfdestruct():
	    self.take_points(self._selfdestruct_cost)
	    self.selfdestruct()
	    return True
	return False
    
    
class Terminator_Marine_with_Assault_Cannon(Terminator_Marine_with_Storm_Bolter, ReloadsMixIn):
    
    type = TERMINATOR_MARINE_WITH_ASSAULT_CANNON
    
    _button_class_list = (  buttons.AttackButton, buttons.ShootButton,\
			    buttons.MovenshootButton, buttons.OverwatchButton,\
                            buttons.UsedoorButton, buttons.AutofireButton, buttons.ReloadButton) #7 actions
    ammo = 10
    _max_fire_bonus = 4
    _autofire_cost = 2
    _autofire_ammo_use = 5


    def tt_text(self):

	text = Terminator_Marine_with_Storm_Bolter.tt_text(self)

	if self.ammoremaining == 0:
	    ammo_string = "None"
	else: 
            ammo_string = str(self.ammoremaining)
        if self.reloadsremaining == 0:
	    r_string = "None"
	else: 
            r_string = str(self.reloadsremaining)

	return text + "\nAmmo: " + ammo_string + "\nReloads: " + r_string


    def fire_at_empty_space(self):
        
        anim.Flash(self, ASSAULT_CANNON)
        if self.ammo:
          self.ammo -= 1
          d1, d2,d3 = textual.DisplayDie(self.team).pips, textual.DisplayDie(self.team).pips,textual.DisplayDie(self.team).pips
          self.shots_fired += 1
          if d1 == d2 == d3 and self.shots_fired >= 10:
              self.malfunction()
              
         
    def malfunction(self):
        
        squares = self.square.get_adjacents()
        for sqr in squares:
            for thing in sqr.piece.sprites():
                if hasattr(thing,"mal_kill_scorereq"):
                    score = textual.DisplayDie().pips
                    if score >= thing.mal_kill_scorereq:
                        thing.kill()
        self.kill()

        
    def __init__(self, square, facing, squadname=""):
        
        Terminator_Marine_with_Storm_Bolter.__init__(self,square,facing,squadname)
        self.shots_fired = 0
        ReloadsMixIn.__init__(self)

        
    def build_visuals(self):
        
        Terminator_Marine_with_Storm_Bolter.build_visuals(self)
        self._clocks.add(textual.AmmoClock(self))

        
    def can_autofire(self):
        """-> int or None
        return cost of auto-firing
        """
        
        if self.ammoremaining >= self._autofire_ammo_use:
            return self._autofire_cost

    
    def goAutofire(self):
        """goAutofire() -> bool
        fire on full auto if possible
        """
        
        cost = self.can_autofire()
        if cost is not None and cost <= self.get_points_avail():
            self.take_points(cost)
            self.autofire()
            return True
        return False
     
     
    def autofire(self):
        """autofire() -> bool
        fire on full auto
        """
        
        anim.Flash(self, ASSAULT_CANNON)
        self.shots_fired += self._autofire_ammo_use
        self.ammoremaining -= self._autofire_ammo_use
        units_in_FIRE_ARC = []
        for unit in self.team.game.pieces.sprites() + self.team.game.features.sprites():
            # We can only destroy closed doors, stealers and marines, which 
            # are in the firearc
            if unit.aauto_kill_scorereq and functions.ang(self.square, unit.square, self.facing) in FIREARC:
                units_in_FIRE_ARC.append(unit)
        while True:
            something_killed = False
            for unit in units_in_FIRE_ARC[:]:
                if functions.seeable(self.square, unit.square):
                    dice = textual.DisplayDie(self.team).pips,\
                    textual.DisplayDie(self.team).pips,\
                    textual.DisplayDie(self.team).pips
                    if max(dice) >= unit.aauto_kill_scorereq:
                        unit.kill()
                        something_killed = True
                        units_in_FIRE_ARC.remove(unit)
            if not(something_killed): break
        self._post_action(AUTOFIRING)

        
    def shoot(self, target):
        """shoot(target) -> None
	fire a shot at target (can be a feature or a piece)"""
	
	def killtarget():
	    "What we do when a kill occurs."
	    # See who has los to the expiring thing, (not incl self)
	    seelist = [m for m in self.team.sprites() if m.can_see(target.square)] #expensive...
	    # Kill the target
	    target.kill(SHOT)
	    # Check the los to blips, but for everybody (stronger than postaction)
	    functions.discover_blips(seelist)
	    print >> DEBUG_STREAM, "Terminator_Marine_with_Assault_Cannon --> shoot --> killtarget" #DEBUG  
	    
	# Go and shoot
      
	# Eyecandy, sounds
	anim.Flash(self, ASSAULT_CANNON)
        self.ammoremaining -= 1
        
	# Roll dice
	fb = int(self._fire_bonus)
        d1, d2, d3 = textual.DisplayDie(self.team).pips, textual.DisplayDie(self.team).pips, textual.DisplayDie(self.team).pips
	print >> DEBUG_STREAM, "Terminator_Marine_with_Assault_Cannon --> shoot, firebonus:%d, dice:%d,%d,%s"%(fb,d1,d2,d3) #DEBUG
	
	# Find the score required to kill
	sr = int(target.aburst_kill_scorereq)
	
        # Try to kill
	if not (self.try_mns or self.overwatching):
	    if self._last_target.has(target): sr -= fb 
	    if d1 >= sr or d2 >= sr or d3 >= sr:
		# Hit:
		killtarget()
                fb = 0
	    else: 
		# Miss:
		fb += 1
		if fb > self._max_fire_bonus: fb = self._max_fire_bonus
		self._last_target.add(target)
        else:
	    if d1 >= sr or d2 >= sr or d3 >= sr: 
		# Hit:
                killtarget()
                fb = 0
	    else: 
		# Miss:
                self._last_target.add(target)
		fb = 0

	# We must only become game.last_actor if we are *acting* (not overwatching)
	if not self.overwatching: 
            self._post_action(SHOOTING, discover=False) # Blip discovery is done by killtarget()

	# Post-_post_action() jiggery pokery
	self._fire_bonus = fb
        
        self.shots_fired += 1
        if d1 == d2 == d3 and self.shots_fired > 10:
            self.malfunction()

        
class Terminator_Marine_with_Chain_Fist(Terminator_Marine_with_Storm_Bolter):
    
    type = TERMINATOR_MARINE_WITH_CHAIN_FIST
    
    _cutdoor_cost = 1
    _button_class_list = (  buttons.AttackButton, buttons.ShootButton,\
			    buttons.MovenshootButton, buttons.OverwatchButton,\
                            buttons.UsedoorButton, buttons.CutDoorButton) #6 actions
    _jammed_button_class_list = (  buttons.AttackButton, buttons.UnjamButton,\
                            buttons.UsedoorButton, buttons.CutDoorButton ) #4 actions
    
    
    def can_cut_door(self, door):
        
        if door.square in self.square.get_adjacents() and self._arc_to(door.square) == F:
            return self._cutdoor_cost 

        
    def cut_door(self, door):
        
        door.kill()
        self.cc_sound.play()
        self._post_action(CUTTING) 

        
    def goCutDoor(self, door):
        """goCutDoor(door) -> bool
        try to cut a door"""
        
        cost = self.can_cut_door(door)
        if cost is not None and self.get_points_avail() >= cost:
            self.take_points(cost)
            self.cut_door(door)
            return True
        else:
            return False

    
class Terminator_Marine_Captain(Terminator_Marine_Sergeant_with_Power_Sword):
    """Good guy with a blingin' power sword and mad powerful grenades.
    He be blowin' up the spot!
    
    Gives +30 seconds to turnclock _and_ +2 CPs.
    
    Attributes:
	cp_bonus
	"""
    _button_class_list = (  buttons.AttackButton, buttons.ShootButton,\
			    buttons.MovenshootButton, buttons.OverwatchButton, buttons.UsedoorButton,buttons.FireGrenadeButton) #6 actions
    _jammed_button_class_list = (  buttons.AttackButton, buttons.UnjamButton, buttons.UsedoorButton,
                                   buttons.FireGrenadeButton) #4 actions

    type = TERMINATOR_MARINE_CAPTAIN
    
    _ccbonusmap = { F:2, L:2, R:2, B:2 }    
    cp_bonus = 2
    ammo = 10
    _grenade_range = 10
    _grenade_cost = 1
    _parry_arc = ( F )


## Grenades

    def can_fire_grenade(self,target_square):
        
        if self.ammoremaining and self._arc_to(target_square) in FIREARC and functions.seeable(self.square, target_square)\
               and (functions.find_range(self, target_square) <= self._grenade_range): 
            return self._grenade_cost

        
    def fire_grenade(self,target_square):
        
        for square in target_square.section.sprites():
             for thing in square.piece.sprites() + square.features.sprites() :
                 if hasattr(thing,"gr_kill_scorereq"):
                     dice = textual.DisplayDie().pips
                     if dice <= thing.gr_kill_scorereq:
                         thing.kill()
        self._post_action(SHOOTING) 

        
    def goFireGrenade(self,target_square):
        
        if self.can_fire_grenade(target_square) and self.get_points_avail >= self._grenade_cost:
            self.take_points(self._grenade_cost)
            self.fire_grenade(target_square)
            
    
class Terminator_Marine_with_Heavy_Flamer(Terminator_Marine_with_Storm_Bolter):
    """Good guy with a napalm weapon.  Scary.  Explode him for hijinks.
    
    Attributes:
	_selfdestruct_cost -- int -- cost of self destruct in APs
	"""
    
    type = TERMINATOR_MARINE_WITH_HEAVY_FLAMER
    
    weapon = FLAMEGUN
    _weapon_range = 12
    _sh_cost = 2
    _selfdestruct_cost = 1
    ammo = 6
    reloads = 0
    _button_class_list = ( buttons.AttackButton, buttons.ShootButton,\
    buttons.SelfdestructButton, buttons.UsedoorButton ) #4 actions
    #DEATHWING reloads optional
    
    keyb = Piece.keyb	    #not inherited from Terminator_Marine_with_Storm_Bolter
    refresh = Piece.refresh #not inherited from Terminator_Marine_with_Storm_Bolter


    def tt_text(self):

	text = Terminator_Marine_with_Storm_Bolter.tt_text(self)

	if self.ammoremaining == 0:
	    ammo_string = "None"
	else: 
            ammo_string = str(self.ammoremaining)

	return text + "\nAmmo: " + ammo_string
    

#### UI

    def build_visuals(self):
	
	Terminator_Marine_with_Storm_Bolter.build_visuals(self)
	# Add an ammo counter
	self._clocks.add(textual.AmmoClock(self))
   

#### Shooting

    def can_shoot(self, targetsquare): #target *must* be a square
        """can_shoot(targetsquare) -> cost / None
	
	If LOS, fire arc, and range permit shooting, returns cost of firing; returns None if impossible.
	Flamers cannot shoot their own section, or squares with closed doors on."""
	
	#not inherited
	for f in targetsquare.features.sprites():
	    if isinstance(f, misc.Door) and f.closed: return
	if self.ammoremaining and not self.square.section.has(targetsquare) and self._arc_to(targetsquare) in FIREARC\
	and functions.seeable(self.square, targetsquare) \
	and functions.find_range(self, targetsquare) <= self._weapon_range:
	    return self._sh_cost
	#hmm... this might be a *better* funct if it took a board *section* as argument and looked for squares in range, etc.
	#that could make UI mucho better anyway. (FUTURE)
	#What about mouseovers?
	
	
    def shoot(self, targetsquare):
        "Flames a section."
	
	#(not inherited)
	anim.Flash(self, FLAMEGUN)
	#self.shootsound.play()
        self.ammoremaining -= 1
	targetsquare.flame() #N.B. even if the section is already flaming, spread and kill things.
	self._post_action(SHOOTING)
	

#### Self-destruction

    def can_selfdestruct(self):
	"""can_selfdestruct() -> int or None
	
	If self-destruct is allowed, return _selfdestruct_cost; else return None."""
	
	if self.ammoremaining: return self._selfdestruct_cost
	
	
    def selfdestruct(self):
	"""selfdestruct() -> None
	
	Kills every piece and feature in own section."""
	
	self.destructsound.play()
        for square in self.square.section.sprites():
            for s in square.piece.sprites(): s.kill(SILENT) #silent kills
            for f in square.features.sprites(): f.kill(SILENT) #even open doors are destroyed
	    square.flame(NOSPREAD) #don't spread the flames, they cover every square in the section already


#### go...()

    def goSelfdestruct(self):
	"""goSelfdestruct() -> bool

	Try to selfdestruct."""
	
	if self.get_points_avail() >= self._selfdestruct_cost and self.can_selfdestruct():
	    self.take_points(self._selfdestruct_cost)
	    self.selfdestruct()
	    return True
	else: return False


    def goHostile(self, target):
	
	# (No inheritance from Terminator_Marine_with_Storm_Bolter)
	if self.goShoot(target.square): return True
	else: return Piece.goHostile(self, target)


    def goSquare(self, square): 
	
	# Target squares rather than pieces
	if Terminator_Marine_with_Storm_Bolter.goSquare(self, square): return True
	else: return self.goShoot(square)

        
class Terminator_Marine_Librarian(Terminator_Marine_with_Storm_Bolter):
    """A good guy who can spend psipoints to increase his close combat bonus
    
    Attributes:
    (instance)
        level               --  int -- from 1 to 4
        psipoints           --  int
        psipointsremaining  --  int
        """

    type = TERMINATOR_MARINE_LIBRARIAN
    
    _ccbonusmap = { F:1, L:1, R:1, B:1 }        

    
    def __init__(self, square, facing, squadname="", level=4):      
        
        self.level = level
        self.psipoints = {1:10, 2:20, 3:30, 4:40}[level]
        self.type = self.type + ", level %d"%level
        self.psipointsremaining = self.psipoints
        Terminator_Marine_with_Storm_Bolter.__init__(self, square, facing, squadname)


    def tt_text(self):

	text = Terminator_Marine_with_Storm_Bolter.tt_text(self)

	if self.psipointsremaining == 0:
	    ammo_string = "None"
	else: 
            ammo_string = str(self.psipointsremaining)

	return text + "\nPsi Points: " + ammo_string
        
        
    def get_cc_scores(self, target):
        
        self.answer = None
        if self.psipointsremaining and self._arc_to(target.square) == F:
            # yes,can spend psipoints
            self.team.player.UIinput.Number_of_Psipoints(self.team.game.phase, self)
            # wait for choice
            while not (self.team.game._done or (self.answer != None)): 
                self.team.game._game_loop()
            self.psipointsremaining -= self.answer
            bonus = self.answer
        else:
            # no,can't spend psipoints
            bonus = 0
        scores1 = Piece.get_cc_scores(self, target)
	return map(lambda x: x + bonus, scores1)
    
    
    def build_visuals(self):
	
	Terminator_Marine_with_Storm_Bolter.build_visuals(self)
	# Add an ammo counter
	self._clocks.add(textual.PsipointsClock(self))
