# Sulk - sprites - toolkit

#*****************************************************************************
#
#                                   Sulk
#
#       Copyright (c) 2002 Toby Woodwark <zirtix@softhome.net>
#
# This program is free software.  See the file COPYING for details.
#
#*****************************************************************************

"""
Sulk dialog graphics.
"""

##############################################################################
####### Import

## Modules
import pygame
from pygame.locals import *

## My modules
import util 
from constants import *  
import graphic


##############################################################################
####### Widgets

class Widget(pygame.sprite.Sprite):
    """Sits in a Dialog (or anything with a 'widgets' attribute).

    Attributes:
    (class)
        clearpack   --  bool    --  if true, leave alone on its row
        decor       --  bool    --  true for special widgets
        click_sound -- a sound (DO NOT USE, sucks)
    (instance)
        parent      --  a WidgetManager
    """

    clearpack = False #default
    decor = False
    
    def __init__(self):
        
        pygame.sprite.Sprite.__init__(self)
        self.parent = None
        
    
    def handle(self, event):
        """handle(event) -> bool
        handle a pygame event"""

        #overload
        return False

    
    def ok(self):
        """-> None
        do any clearing-up / callback before the widgetmanager exits successfully"""

        #overload
        pass


    def kill(self):
        
        pygame.sprite.Sprite.kill(self)
        # Break cycle -- TODO weakref
        if hasattr(self, 'parent'):
            del self.parent


class HorizRuleWidget(Widget):
    """Horiz rule to break up dialogs
    
    We promise not to draw ourselves too big, as the widget is 
    'drawn last' i.e. after packing"""

    clearpack = True
    decor = True #flag for weirdness in Dialog

    def __init__(self, height=1):

        # Will be as wide as the widest widget in the list, with a default
        # and minimum of 3 * PAD.  Height of 1 px default.

        assert type(height) is type(1)

        self.height = height

        Widget.__init__(self)
        self.rect = Rect(0, 0, 3 * PAD, height) #may change
    
    
    def draw_finally(self, width):
        
        image = pygame.Surface((width, self.height))
        image.fill(DIALOG_TEXT_COLOR) #?
        mtop = self.rect.midtop
        self.image = image
        self.rect = image.get_rect()
        self.rect.midtop = mtop

    
class CRTextWidget(Widget):
    """Just shows text.

    Takes text with carriage-returns.
    """


    def __init__(self, text):
        
        Widget.__init__(self)
        
        # Render the text
        self.image = util.render.para_with_CR(DIALOG_FONT, text, False, True, DIALOG_TEXT_COLOR)
        self.rect = self.image.get_rect()

    
    def set_text(self, text):
        """set_text(text) -> None
        change the text of the widget in place"""
        #use with care, the size of the dialog will not change.
        
        self.image = util.render.para_with_CR(DIALOG_FONT, text, False, True, DIALOG_TEXT_COLOR)
        self.rect = self.image.get_rect()


class MTextDisplayWidget(Widget):
    """Shows title text.

    Takes a textarray for util.render.mtext
    """
    
    clearpack = True
    
    
    def __init__(self, textarray):
        
        Widget.__init__(self)
        
        # Render the text
        self.image = util.render.mtext(DIALOG_FONT, textarray, DIALOG_MAX_QUERY_WIDTH, 1, DIALOG_TEXT_COLOR)
        self.rect = self.image.get_rect()


class ButtonWidget(Widget):
    """Non-togglable button.
    
    Attributes:
    (instance)
        callback                -- a function of no arguments
        kill_dialog_on_callback -- bool -- true if the dialog and its handler should die on callback
        mnem                    -- a length-1 string, or None
        text                    -- a string
        """

    def __init__(self, text, mnem, callback, kill_dialog_on_callback=False):
        
        assert callable(callback)
        
        Widget.__init__(self)
        self.text, self.callback, self.kill_dialog_on_callback = \
        text, callback, kill_dialog_on_callback

        self.mnem = mnem.lower()
        
        # Render
        textimage = util.render.with_mnemonic(
        CLICKBOX_FONT,
        " %s " % text,
        mnem,
        1,
        CLICKBOX_TEXT_COLOR
        )
        
        self.image = util.draw.borderize(textimage, HALF, CLICKBOX_BACK_COLOR, CLICKBOX_EDGE_COLOR)
        self.rect = self.image.get_rect()
            
    
    def hit(self):

        self.click_sound.play() #sucks
        if self.kill_dialog_on_callback:
            self.parent.kill()
        else:
            self.parent.update()
        self.callback()


    def handle(self, event):

        if (event.type == MOUSEBUTTONUP and self.rect.collidepoint(event.pos)\
        and event.button in M__SELECT)\
        or (event.type == KEYDOWN and event.unicode.lower() == self.mnem):
            self.hit()
            return True
        return False

        
class CheckboxWidget(ButtonWidget): #IN PROGRESS
    """Togglable button.
    
    Attributes:
    (class)
        boximg                  -- dict
    (instance)
        checked                 -- bool
        mnem                    -- a length-1 string, or None
        text                    -- a string
        okfunc                  -- a function of boolean argument, called by ok()
        """

    def __init__(self, text, mnem, okfunc, checked=False):
        
        Widget.__init__(self)
        
        self.text   = text
        self.mnem   = mnem.lower()
        self.checked = checked
        self.okfunc = okfunc

        self.draw()
        self.rect   = self.image.get_rect()


    def draw(self):
        
        # Get box image
        if self.checked:
            boximage = self.boximg[PRESSED]
        else:
            boximage = self.boximg[ENABLED]

        boxrect = self.boximg[PRESSED].get_rect()
        
        # Render text
        textimage = util.render.with_mnemonic(
        CLICKBOX_FONT,
        " %s " % self.text,
        self.mnem,
        1,
        CLICKBOX_TEXT_COLOR
        )
            
        textrect = textimage.get_rect()

        textrect.midleft = boxrect.midright # padding?

        # Make image
        newrect = textrect.union(boxrect)
        img = pygame.Surface(newrect.size)
        img.blit(textimage, textrect)
        img.blit(boximage, boxrect)
        
        self.image = util.draw.borderize(img, HALF, CLICKBOX_BACK_COLOR, CLICKBOX_EDGE_COLOR)
            

    def hit(self):
        
        self.click_sound.play()
        self.checked = (not self.checked)
        self.draw()
        self.parent.update() # to redraw whole dialog (blech)


    def ok(self):

        self.okfunc(self.checked)


class RadioButtonWidget(ButtonWidget):
    """Switchable button.
    
    Attributes:
    (class)
        boximg                  -- dict of int:surf
    (instance)
        funcdict                -- dict of (text,mnem):okfunc
                                   (okfunc is called by ok(), with no arguments)
        selected                -- key to funcdict
        _areas                  -- dict of (text,mnem):rect
        """

    #Should this be a widgetmanager/widget mixin instead? - TW

    def __init__(self, funcdict, selected=None):

        assert funcdict
        
        Widget.__init__(self)
        
        # Get a default if none supplied
        if not selected: 
            selected = funcdict.keys()[0]
        self.selected = selected
        self.funcdict = funcdict

        self.draw()
        self.rect = self.image.get_rect()


    def draw(self):  #cycle wastage here - TW

        #TODO -- subsurfaces?

        rect = Rect(util.zerorect)
        boxkeys = self.funcdict.keys()
        boxkeys.sort()
        images = {}
        self._areas = {}

        y = 0

        for k in boxkeys:
            
            # Get box image
            if self.selected is k:
                boximage = self.boximg[PRESSED]
            else:
                boximage = self.boximg[ENABLED]
                
            # Render text
            textimage = util.render.with_mnemonic(
            CLICKBOX_FONT,
            " %s " % k[0],
            k[1],
            1,
            CLICKBOX_TEXT_COLOR
            )
                
            textrect = textimage.get_rect()
            boxrect = boximage.get_rect()
            textrect.midleft = boxrect.midright # padding?

            # Make image
            newrect = boxrect.union(textrect)
            img = pygame.Surface(newrect.size)
            img.blit(textimage, textrect)
            img.blit(boximage, boxrect)
            
            images[k] = img

            # Resize the containing rect
            newrect.top = y
            rect.height += newrect.height
            y += newrect.height
            self._areas[k] = newrect
            rect.width = max(newrect.width, rect.width)

        image = pygame.Surface(rect.size)
        image.fill(CLICKBOX_BACK_COLOR)
        
        for k in boxkeys:
            image.blit(images[k], self._areas[k])
            
        self.image = util.draw.borderize(image, HALF, CLICKBOX_BACK_COLOR, CLICKBOX_EDGE_COLOR)
        
            
    def hit(self, selected):

        self.click_sound.play()
        self.selected = selected
        self.draw()
        self.parent.update()
        
        
    def handle(self, event):
        
        d = self.funcdict
        
        if (event.type == MOUSEBUTTONUP and self.rect.collidepoint(event.pos)\
        and event.button in M__SELECT):
            #Mung the pos to make it match _areas dict
            relpos = event.pos[0] - self.rect.left - HALF/2, event.pos[1] - self.rect.top - HALF/2
            # Find the hit
            for k in self._areas.keys():
                if self._areas[k].collidepoint(relpos):
                    self.hit(k)
                    return True
        elif event.type == KEYDOWN:
            for k in d.keys():
                if event.unicode.lower() == k[1].lower():
                    self.hit(k)
                    return True
        return False


    def ok(self):

        self.funcdict[self.selected]()


##############################################################################
####### Dialog

class WidgetManagerMixIn:
    """Container.  Parent class
    
    """


    def __init__(self, *widgets):

        self.widgets = []
        self.packed = False
        for widget in widgets:
            self.add_widget(widget)

    
    def add_widget(self, widget):

        if self.packed:
            raise UserWarning, "already packed"
        else:
            widget.parent = self
            if widget not in self.widgets:
                self.widgets.append(widget)
        
        
    def handle(self, event):

        for widget in self.widgets:
            if widget.handle(event):
                return True
        return False
        
    
    def pack(self): 
        
        #overload
        self.packed = True
        

    def kill(self):

        for widget in self.widgets:
            widget.kill()


    def ok(self):

        for widget in self.widgets:
            widget.ok()
 

class Dialog(graphic.Graphic, WidgetManagerMixIn):
    """A dialog window.
   
   Attributes:
    (class)
       all_dialogs  -- a spritegroup -- all the dialogs, maintained by SulkGame
       """
    
    render_clip = None
    render_layer = 10
    
    tt_text = lambda self: ""


    def __init__(self, *widgets):

        #TODO fade in, drop in, or do some other funky stuff?
        
        graphic.Graphic.__init__(self, self.all_dialogs)
        WidgetManagerMixIn.__init__(self, *widgets)
        
        
    def pack(self):
        
        # Put widgets into layout rows
        rows = [[]]
        for widget in self.widgets:
            #hmm
            if getattr(widget, 'clearpack', 0):
                if rows[-1]:
                    rows.append([widget])
                else:
                    rows[-1] = [widget]
                rows.append([])
            elif widget.rect.width\
            + reduce(lambda x, y: x + PAD + y, map(lambda x: x.rect.width, rows[-1]), 0)\
            + 2 * PAD < MAX_DIALOG_WIDTH:
                rows[-1].append(widget)
            else:
                rows.append([widget])
        
        # Make a rect big enough for all the widgets
        rect = Rect(0,0,PAD,PAD)
        y = 0
        for row in rows:
            # TODO -- centered rows?
            y += HALF
            x = PAD #?
            maxbottom = 0
            for widget in row:
                widget.rect.topleft = x, y
                x = widget.rect.right + PAD
                maxbottom = max(maxbottom, widget.rect.bottom + HALF) #? HALF?
                rect.union_ip(widget.rect) #?
            y = maxbottom

        rect.height += PAD
        rect.width += PAD
        
        # Adjust clearpack widgets to be centered, all others remain left-justified (FIXME)
        for widget in self.widgets:
            if widget.clearpack:
                widget.rect.centerx = rect.centerx
        
        # and draw the decorative rules
            if widget.decor:
                widget.draw_finally(rect.width - 2 * PAD) #hmm
        
        # Make an image
        image = pygame.Surface(rect.size)
        image.fill(DIALOG_BACK_COLOR)
        pygame.draw.rect(image, DIALOG_EDGE_COLOR, rect, 1)
        image.set_alpha(DIALOG_ALPHA)
            
        # Move to the middle of the screen 
        oldleft, oldtop = rect.topleft
        rect = image.get_rect()
        rect.center = DIALOG_CENTER
        rect.clamp_ip(SCREEN_RECT.inflate(-4 * PAD, -4 * PAD))

        # Keep track of my position change (see handle)
        self.deltax, self.deltay = rect.left - oldleft, rect.top - oldtop #THIS WORKS -TW

        # Bind attrs
        self.rect = rect
        self.base_image = image

        WidgetManagerMixIn.pack(self)
        self.update()


    def update(self): #FIXME heavy.... -TW
        
        self.image = self.base_image.convert() #a copy
        for widget in self.widgets:
            self.image.blit(widget.image, widget.rect)


    def handle(self, event):

        if getattr(event, 'pos', False):
            # Mung the pos to get the right results in widget.handle
            neweventdict = dict(event.dict)
            #print >> DEBUG_STREAM, "NewDialog --> handle, event.dict:", event.dict #DEBUG
            neweventdict['pos'] = (event.pos[0] - self.deltax, event.pos[1] - self.deltay)
            #print >> DEBUG_STREAM, "NewDialog --> handle, new event.dict:", neweventdict #DEBUG
            event = pygame.event.Event(event.type, neweventdict) #KLUDGE
        if WidgetManagerMixIn.handle(self, event):
            return True
        return False


    def kill(self):

        WidgetManagerMixIn.kill(self)
        graphic.Graphic.kill(self)

NewDialog = Dialog #DEPRECATED
