# Sulk - util - render

#*****************************************************************************
#
#                                   Sulk
#
#       Copyright (c) 2002 Toby Woodwark <zirtix@softhome.net>
#
# This program is free software.  See the file COPYING for details.
#
#*****************************************************************************

"""
Sulk font rendering functions.
"""

##############################################################################
####### Import

import pygame

from constants import DEBUG_STREAM


##############################################################################
####### Functions 

def with_mnemonic(font, text, letterstring, antialias, fore_RGBA, back_RGBA=None):
    """util.render.with_mnemonic(font, text, letterstring, antialias, fore_RGBA, back_RGBA=None) -> Surface instance
    render 'text' with the first occurrence of 'letterstring' underlined
    
    We intelligently guess that if an occurrence of the letter is at the
    start of a word and/or is uppercase, we should prefer that occurrence for
    the mnemonic one.
    
    If the letter can't be found, we return a normal render."""

    letter = letterstring[0]
    
    # Find occurrences of letter
    upperletter, lowerletter = letter.upper(), letter.lower()
    indexupper, indexlower = [], []
    for i in range(len(text)):
	if text[i] == upperletter: indexupper.append(i)
	elif text[i] == lowerletter: indexlower.append(i)
    
    # Find the best occurrence of letter
    for i in indexupper + indexlower: #priority to upperletter
	if i == 0 or text[i-1].isspace():
	    #we're at the beginning of a word
	    letterindex = i 
	    letter = text[i]
	    break
    else:
	for i in indexupper + indexlower:
	    letterindex = i
	    break
	else:
            # Letter is absent
	    print >> DEBUG_STREAM,\
            "Error: util.fontutil.render_with_mnemonic:\
            letter '%s' not in '%s'"%(letter, text) #DEBUG
            return font.render(text, antialias, fore_RGBA, back_RGBA)
    
    # Split the text up
    beforetext, aftertext = text[0:letterindex], text[letterindex+1:len(text)]
    
    #print >> DEBUG_STREAM, \
    #"render_with_mnemonic--> beforetext,letter,aftertext:", beforetext,letter,aftertext #DEBUG

    # Render three surfaces
    font.set_underline(1)
    lettersurface = font.render(letter, antialias, fore_RGBA)
    font.set_underline(0)
    beforesurface = font.render(beforetext, antialias, fore_RGBA)
    aftersurface = font.render(aftertext, antialias, fore_RGBA)

    # Create the eventual surface
    rect1, rect2, rect3 = beforesurface.get_rect(), lettersurface.get_rect(), aftersurface.get_rect()
    width = rect1.width + rect2.width + rect3.width
    height = max([rect1.height, rect2.height, rect3.height])
    productsurface = pygame.Surface((width, height), 0, lettersurface) #hmm
    if back_RGBA: 
        productsurface.fill(back_RGBA)

    # Blit them on
    productsurface.blit(beforesurface, (0,0))
    productsurface.blit(lettersurface, (rect1.width,0))
    productsurface.blit(aftersurface, (rect1.width + rect2.width,0))

    return productsurface
    

def _string_to_lines_by_max_chars(text, length):
    """_string_to_lines_by_max_chars(text, length) -> list
    format a paragraph
    
    Given string 'text', returns a list of lines no more than int 'length'
    characters long."""

    words = text.split()
    lines = []
    line = ''
    while words:
	word = words.pop(0)
	if len(word) + 1 + len(line) > length: 
	    lines.append(line)
	    line = word
	else:
	    if line:
		line = " ".join((line, word))
	    else: line = word
    else:
	#catch the last line
	lines.append(line)
    return lines


def para_with_CR(font, text, centered, antialias, fore_RGBA, back_RGBA=None):
    """util.render.para_with_CR(font, lines, centered, antialias, fore_RGBA, back_RGBA=None) -> Surface instance
    render a string containing newlines as a paragraph"""
    
    lines = text.split("\n")
    # Turn empty lines into single spaces
    lines = map(lambda s: (s or " "), lines)

    return para(font, lines, centered, antialias, fore_RGBA, back_RGBA)

    
def para(font, lines, centered, antialias, fore_RGBA, back_RGBA=None):
    """util.render.para(font, lines, centered, antialias, fore_RGBA, back_RGBA=None) -> Surface instance
    render list of strings as a paragraph
    
    If 'centered' is true, render lines centered.  Otherwise left-justify."""
    
    # Prepare to make surfaces, one per line
    productwidth = 0
    productheight = 0
    lsurfaces = []
    
    # Render surfaces and find size
    for line in lines:
	s = font.render(line, antialias, fore_RGBA)
	lsurfaces.append(s)
	productwidth = max(s.get_rect().width, productwidth)
	productheight += s.get_rect().height

    # Start final surface
    if back_RGBA: 
	productsurface = pygame.Surface((productwidth, productheight))
	productsurface.fill(back_RGBA)
    else:
	productsurface = pygame.Surface((productwidth, productheight), 0, lsurfaces[0]) 
	#(don't know why, but this works for alpha)
    
    # Blit the lines on in order
    y = 0
    if centered:
	for s in lsurfaces:
	    r = s.get_rect()
	    r.centerx = productsurface.get_rect().centerx #centered text
	    r.top = y
	    productsurface.blit(s, r)
	    y += r.height
    else:
	for s in lsurfaces:
	    productsurface.blit(s, (0, y))
	    y += s.get_rect().height
	
    return productsurface

    
def mtext(font, mtextarray, maxwidth, antialias, fore_RGBA, back_RGBA=None):
    """util.render.mtext(font, mtextarray, maxwidth, antialias, fore_RGBA, back_RGBA=None) -> Surface instance
    render long text in multiple lines

    mtextarray	-- 2-tuple of a string and a tuple of strings.  The string is 
		   rendered as a centered heading and the tuple elements 
		   (if any) are each rendered as a paragraph.
    antialias	-- bool
    maxwidth	-- int -- width in pixels."""
	
    # Work out dimensions
    avg_letter_width = int(font.size("AbCdEfGhIjKlMnOpQrStUvWxYz :,.")[0] / 30) #kludge
    max_letters_per_line = int(maxwidth / avg_letter_width)
    # print >> DEBUG_STREAM, 'render_mtext --> max_letters_per_line:', max_letters_per_line #DEBUG
    
    # Get the strings
    headerstring = mtextarray[0]
    pstringlist = list(mtextarray[1])
    
    # Loop through making lines
    headerlines = _string_to_lines_by_max_chars(headerstring, max_letters_per_line)
    plinesarray = []
    for p in pstringlist:
	plinesarray.append(_string_to_lines_by_max_chars(p, max_letters_per_line))
	
    # Now make surfaces, one per line
    productwidth = productheight = 0
    headersurfaces = []
    psurfaces = []
    for line in headerlines:
	#print >> DEBUG_STREAM, 'render_mtext --> headerline:', line #DEBUG
	s = font.render(line, antialias, fore_RGBA) #comes with SRCALPHA (what? -- TW)
	headersurfaces.append(s)
	if s.get_rect().width > productwidth: productwidth = s.get_rect().width
	productheight += s.get_rect().height
    for plist in plinesarray:
	productheight += font.get_linesize()
	for line in plist:
	    #print >> DEBUG_STREAM, 'render_mtext --> pline:', line #DEBUG
	    s = font.render(line, antialias, fore_RGBA)
	    psurfaces.append(s)
	    if s.get_rect().width > productwidth: productwidth = s.get_rect().width
	    productheight += s.get_rect().height
	    
    # Start final surface
    productsurface = pygame.Surface((productwidth, productheight), 0, headersurfaces[0]) #is this is the only way to get SRCALPHA on this surface?!  HELP!  Am I an idiot? (eh?)
    if back_RGBA: productsurface.fill(back_RGBA)
    
    # Blit the lines on in order
    y = 0
    for s in headersurfaces:
	r = s.get_rect()
	r.centerx = productsurface.get_rect().centerx #centered text
	r.top = y
	productsurface.blit(s, r)
	y += r.height
    for plist in plinesarray: 
	y += font.get_linesize()
	for line in plist:
	    s = psurfaces.pop(0)
	    productsurface.blit(s, (0, y))
	    y += s.get_rect().height
    return productsurface
